# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos, pi, sin

from wulfric.cell._sc_constructors import (
    BCC,
    BCT,
    CUB,
    FCC,
    HEX,
    MCL,
    MCLC,
    ORC,
    ORCC,
    ORCF,
    ORCI,
    RHL,
    TET,
    TRI,
)
from wulfric.constants._numerical import TORADIANS
from wulfric.constants._sc_notation import BRAVAIS_LATTICE_VARIATIONS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_cell_example(lattice_variation: str = None, convention: str = "sc"):
    r"""
    Examples of the Bravais lattices as defined in the paper by Setyawan and Curtarolo [1]_.

    Parameters
    ----------
    lattice_variation : str, optional
        Name of the lattice type or variation to be returned. For available names see
        documentation of each :ref:`user-guide_conventions_bravais-lattices`.
        Case-insensitive.
    convention : str, default "sc"
        Name of the convention that is used for cell standardization. Case-insensitive.
        Supported conventions are

        * "sc" - for Setyawan and Curtarolo [1]_.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a direct cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    Raises
    ------
    ValueError
        If ``convention`` is not supported.

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_cell_example("cub")
        array([[3.14159265, 0.        , 0.        ],
               [0.        , 3.14159265, 0.        ],
               [0.        , 0.        , 3.14159265]])
        >>> wulf.cell.get_cell_example("ORCF3")
        array([[0.        , 1.96349541, 2.61799388],
               [1.57079633, 0.        , 2.61799388],
               [1.57079633, 1.96349541, 0.        ]])
    """

    convention = convention.lower()

    if convention != "sc":
        raise ValueError(
            f'"{convention}" convention is not supported. Supported is "sc".'
        )

    correct_inputs = set(map(lambda x: x.lower(), BRAVAIS_LATTICE_VARIATIONS)).union(
        set(
            map(
                lambda x: x.translate(str.maketrans("", "", "12345ab")).lower(),
                BRAVAIS_LATTICE_VARIATIONS,
            )
        )
    )

    if (
        not isinstance(lattice_variation, str)
        or lattice_variation.lower() not in correct_inputs
    ):
        message = f"There is no {lattice_variation} Bravais lattice. Available examples are:\n"
        for name in BRAVAIS_LATTICE_VARIATIONS:
            message += f"  * {name}\n"
        raise ValueError(message)

    lattice_variation = lattice_variation.lower()

    if lattice_variation == "cub":
        cell = CUB(pi)
    elif lattice_variation == "fcc":
        cell = FCC(pi)
    elif lattice_variation == "bcc":
        cell = BCC(pi)
    elif lattice_variation == "tet":
        cell = TET(pi, 1.5 * pi)
    elif lattice_variation in ["bct1", "bct"]:
        cell = BCT(1.5 * pi, pi)
    elif lattice_variation == "bct2":
        cell = BCT(pi, 1.5 * pi)
    elif lattice_variation == "orc":
        cell = ORC(pi, 1.5 * pi, 2 * pi)
    elif lattice_variation in ["orcf1", "orcf"]:
        cell = ORCF(0.7 * pi, 5 / 4 * pi, 5 / 3 * pi)
    elif lattice_variation == "orcf2":
        cell = ORCF(1.2 * pi, 5 / 4 * pi, 5 / 3 * pi)
    elif lattice_variation == "orcf3":
        cell = ORCF(pi, 5 / 4 * pi, 5 / 3 * pi)
    elif lattice_variation == "orci":
        return ORCI(pi, 1.3 * pi, 1.7 * pi)
    elif lattice_variation == "orcc":
        cell = ORCC(pi, 1.3 * pi, 1.7 * pi)
    elif lattice_variation == "hex":
        cell = HEX(pi, 2 * pi)
    elif lattice_variation in ["rhl1", "rhl"]:
        # If alpha = 60 it is effectively FCC!
        cell = RHL(pi, 70)
    elif lattice_variation == "rhl2":
        cell = RHL(pi, 110)
    elif lattice_variation == "mcl":
        cell = MCL(pi, 1.3 * pi, 1.6 * pi, alpha=75)
    elif lattice_variation in ["mclc1", "mclc"]:
        cell = MCLC(pi, 1.4 * pi, 1.7 * pi, 80)
    elif lattice_variation == "mclc2":
        cell = MCLC(1.4 * pi * sin(75 * TORADIANS), 1.4 * pi, 1.7 * pi, 75)
    elif lattice_variation == "mclc3":
        b = pi
        x = 1.1
        alpha = 78
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha) * 1.8
        a = x * b * sin(ralpha)
        cell = MCLC(a, b, c, alpha)
    elif lattice_variation == "mclc4":
        b = pi
        x = 1.2
        alpha = 65
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha)
        a = x * b * sin(ralpha)
        cell = MCLC(a, b, c, alpha)
    elif lattice_variation == "mclc5":
        b = pi
        x = 1.4
        alpha = 53
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha) * 0.9
        a = x * b * sin(ralpha)
        cell = MCLC(a, b, c, alpha)
    elif lattice_variation in ["tri1a", "tri1", "tri", "tria"]:
        cell = TRI(1, 1.5, 2, 120, 110, 100, input_reciprocal=True)
    elif lattice_variation in ["tri2a", "tri2"]:
        cell = TRI(1, 1.5, 2, 120, 110, 90, input_reciprocal=True)
    elif lattice_variation in ["tri1b", "trib"]:
        cell = TRI(1, 1.5, 2, 60, 70, 80, input_reciprocal=True)
    elif lattice_variation == "tri2b":
        cell = TRI(1, 1.5, 2, 60, 70, 90, input_reciprocal=True)

    return cell


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


R"""Syntax sugar"""

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class _SyntacticSugar(dict):
    r"""
    Syntactic sugar for any dictionary.

    This class does only one thing. It allows to write
    ``atoms.names`` instead of ``atoms["names"]``.

    Examples
    --------

    Two code examples below give equivalent result

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = wulf.Atoms()
        >>> atoms.names = ["Cr1", "Cr2"]
        >>> atoms.positions = [[0, 0, 0], [0.5, 0.5, 0.5]]
        {'names': ['Cr1', 'Cr2'], 'positions': [[0, 0, 0], [0.5, 0.5, 0.5]]}

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {}
        >>> atoms["names"] = ["Cr1", "Cr2"]
        >>> atoms["positions"] = [[0, 0, 0], [0.5, 0.5, 0.5]]
        {'names': ['Cr1', 'Cr2'], 'positions': [[0, 0, 0], [0.5, 0.5, 0.5]]}

    ``Atom`` class behaves as dictionary

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = wulf.Atoms()
        >>> atoms.names = ["Cr1", "Cr2"]
        >>> atoms.positions = [[0, 0, 0], [0.5, 0.5, 0.5]]
        {'names': ['Cr1', 'Cr2'], 'positions': [[0, 0, 0], [0.5, 0.5, 0.5]]}

    """

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def add_sugar(dictionary: dict) -> _SyntacticSugar:
    r"""
    Takes any dictionary and add attribute-like access to the key-values to it.

    Parameters
    ----------
    dictionary : dict
        Dictionary for the addition of the syntax sugar.

    Returns
    -------
    candy : :py:class:`._SyntacticSugar`
        Same dictionary with the easy access to the key-value pairs.

    Raises
    ------
    ValueError
        If ``not isinstance(dictionary, dict)``.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {"names" : ["Cr1", "Cr2"]}
        >>> atoms.names
        Traceback (most recent call last):
        ...
        AttributeError: 'dict' object has no attribute 'names'
        >>> atoms = wulf.add_sugar(atoms)
        >>> atoms.names
        ['Cr1', 'Cr2']
        >>> atoms.positions = [[0, 0, 0], [0.5, 0.5, 0.5]]
        >>> atoms.positions
        [[0, 0, 0], [0.5, 0.5, 0.5]]
        >>> # Note that it still behaves as a dictionary
        >>> atoms["positions"] = [[0.5, 0.5, 0.5], [0, 0, 0]]
        >>> atoms.positions
        [[0.5, 0.5, 0.5], [0, 0, 0]]
        >>> atoms["positions"]
        [[0.5, 0.5, 0.5], [0, 0, 0]]
        >>> atoms
        {'names': ['Cr1', 'Cr2'], 'positions': [[0.5, 0.5, 0.5], [0, 0, 0]]}

    """

    if not isinstance(dictionary, dict):
        raise ValueError(
            f"dictionary should be an instance of python dict, got {type(dictionary)}."
        )

    candy = _SyntacticSugar()

    for key in dictionary:
        candy[key] = dictionary[key]

    return candy


def remove_sugar(candy: dict) -> dict:
    r"""
    Takes any dictionary and remove attribute-like access to the key-values to it.

    Parameters
    ----------
    candy : dict
        Dictionary for the addition of the syntax sugar.

    Returns
    -------
    dictionary : dict
        Same dictionary without the easy access to the key-value pairs.

    Raises
    ------
    ValueError
        If ``not isinstance(candy, dict)``.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {"names" : ["Cr1", "Cr2"]}
        >>> atoms = wulf.add_sugar(atoms)
        >>> atoms.names
        ['Cr1', 'Cr2']
        >>> atoms.positions = [[0, 0, 0], [0.5, 0.5, 0.5]]
        >>> atoms = wulf.remove_sugar(atoms)
        >>> atoms.names
        Traceback (most recent call last):
        ...
        AttributeError: 'dict' object has no attribute 'names'
        >>> atoms
        {'names': ['Cr1', 'Cr2'], 'positions': [[0, 0, 0], [0.5, 0.5, 0.5]]}

    """

    if not isinstance(candy, dict):
        raise ValueError(
            f"candy should be an instance of python dict, got {type(dictionary)}."
        )

    dictionary = {}

    for key in candy:
        dictionary[key] = candy[key]

    return dictionary


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos, sqrt

import numpy as np

from wulfric._numerical import compare_numerically
from wulfric.constants._numerical import TODEGREES, TORADIANS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_volume(*args):
    r"""
    Computes volume.

    Three types of arguments are expected:

    * One argument.
        Matrix ``cell``.
        Volume is computed as:

        .. math::
            V = \boldsymbol{v_1} \cdot (\boldsymbol{v_2} \times \boldsymbol{v_3})
    * Three arguments.
        Vectors ``v1``, ``v2``, ``v3``.
        Volume is computed as:

        .. math::
            V = \boldsymbol{v_1} \cdot (\boldsymbol{v_2} \times \boldsymbol{v_3})
    * Six arguments.
        Parameters ``a``, ``b``, ``c``, ``alpha``, ``beta``, ``gamma``.
        Volume is computed as:

        .. math::
            V = abc\sqrt{1+2\cos(\alpha)\cos(\beta)\cos(\gamma)-\cos^2(\alpha)-\cos^2(\beta)-\cos^2(\gamma)}


    Parameters
    ----------
    v1 : (3,) |array-like|_
        First vector.
    v2 : (3,) |array-like|_
        Second vector.
    v3 : (3,) |array-like|_
        Third vector.
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    a : float, default 1
        Length of the :math:`v_1` vector.
    b : float, default 1
        Length of the :math:`v_2` vector.
    c : float, default 1
        Length of the :math:`v_3` vector.
    alpha : float, default 90
        Angle between vectors :math:`v_2` and :math:`v_3`. In degrees.
    beta : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_3`. In degrees.
    gamma : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_2`. In degrees.

    Returns
    -------
    volume : float
        Volume of corresponding region in space.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.geometry.get_volume([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        1.0
        >>> wulf.geometry.get_volume([1, 0, 0], [0, 1, 0], [0, 0, 1])
        1.0
        >>> wulf.geometry.get_volume(1, 1, 1, 90, 90, 90)
        1.0
    """

    if len(args) == 1:
        cell = np.array(args[0])
    elif len(args) == 3:
        cell = np.array(args)
    elif len(args) == 6:
        a, b, c, alpha, beta, gamma = args
        alpha = alpha * TORADIANS
        beta = beta * TORADIANS
        gamma = gamma * TORADIANS
        sq_root = (
            1
            + 2 * cos(alpha) * cos(beta) * cos(gamma)
            - cos(alpha) ** 2
            - cos(beta) ** 2
            - cos(gamma) ** 2
        )
        return a * b * c * sqrt(sq_root)
    else:
        raise ValueError(
            "Unable to identify input. "
            + "Supported: one (3, 3) array-like, or three (3,) array-like, or 6 floats."
        )

    return float(abs(np.linalg.det(cell)))


def get_angle(v1, v2, radians=False):
    r"""
    Computes angle between two vectors.

    .. math::

        \alpha
        =
        \dfrac{(\boldsymbol{v_1} \cdot \boldsymbol{v_2})}
        {\vert\boldsymbol{v_1}\vert\cdot\vert\boldsymbol{v_2}\vert}

    Parameters
    ----------
    v1 : (3,) |array-like|_
        First vector.
    v2 : (3,) |array-like|_
        Second vector.
    radians : bool, default False
        Whether to return value in radians.

    Returns
    -------
    angle: float
        Angle in degrees or radians (see ``radians``).

    Raises
    ------
    ValueError
        If one of the vectors is zero vector (or both). Norm is compared against
        :numpy:`finfo`\ (float).eps.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.geometry.get_angle([1, 0, 0], [0, 0, 1])
        90.0
        >>> wulf.geometry.get_angle([1, 0, 0], [1, 0, 0])
        0.0
        >>> round(wulf.geometry.get_angle([1, 0, 0], [1, 1, 1]), 4)
        54.7356
    """

    # Normalize vectors
    v1_norm = np.linalg.norm(v1)
    v2_norm = np.linalg.norm(v2)
    if abs(v1_norm) <= np.finfo(float).eps or abs(v2_norm) <= np.finfo(float).eps:
        raise ValueError("Angle is ill defined (zero vector).")

    v1 = np.array(v1) / v1_norm
    v2 = np.array(v2) / v2_norm

    alpha = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
    if radians:
        return float(alpha)

    return float(alpha * TODEGREES)


def parallelepiped_check(
    a,
    b,
    c,
    alpha,
    beta,
    gamma,
    raise_error=False,
    length_tolerance=1e-8,
    angle_tolerance=1e-4,
):
    r"""
    Checks if parallelepiped is valid.

    The conditions are

    * :math:`a > 0`
    * :math:`b > 0`
    * :math:`c > 0`
    * :math:`0 < \alpha < 180`
    * :math:`0 < \beta < 180`
    * :math:`0 < \gamma < 180`
    * :math:`\gamma < \alpha + \beta < 360 - \gamma`
    * :math:`\beta < \alpha + \gamma < 360 - \beta`
    * :math:`\alpha < \beta + \gamma < 360 - \alpha`

    Parameters
    ----------
    a : float
        Length of the :math:`\boldsymbol{v_1}` vector.
    b : float
        Length of the :math:`\boldsymbol{v_2}` vector.
    c : float
        Length of the :math:`\boldsymbol{v_3}` vector.
    alpha : float
        Angle between vectors :math:`\boldsymbol{v_2}` and :math:`\boldsymbol{v_3}`. In degrees.
    beta : float
        Angle between vectors :math:`\boldsymbol{v_1}` and :math:`\boldsymbol{v_3}`. In degrees.
    gamma : float
        Angle between vectors :math:`\boldsymbol{v_1}` and :math:`\boldsymbol{v_2}`. In degrees.
    raise_error : bool, default False
        Whether to raise error if parameters can not form a parallelepiped.
    length_tolerance : float, default :math:`10^{-8}`
        Numerical tolerance for the length variables. Default value is chosen in the
        contexts of condense matter physics, assuming that length is given in Angstroms.
        Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Numerical tolerance for the angle variables. Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    result: bool
        Whether the parameters could from a parallelepiped.

    Raises
    ------
    ValueError
        If parameters can not form a parallelepiped.
        Only raised if ``raise_error`` is ``True`` (it is ``False`` by default).

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.geometry.parallelepiped_check(1, 1, 1, 90, 90, 90)
        True
        >>> wulf.geometry.parallelepiped_check(1, 1, 1, 30, 20, 110)
        False
        >>> wulf.geometry.parallelepiped_check(1, -1, 1, 90, 90, 90)
        False
        >>> wulf.geometry.parallelepiped_check(1, 0, 1, 90, 90, 90)
        False
        >>> wulf.geometry.parallelepiped_check(1, 1, 1, 90, 199, 90)
        False

    """

    result = (
        compare_numerically(a, ">", 0.0, eps=length_tolerance)
        and compare_numerically(b, ">", 0.0, eps=length_tolerance)
        and compare_numerically(c, ">", 0.0, eps=length_tolerance)
        and compare_numerically(alpha, "<", 180.0, eps=angle_tolerance)
        and compare_numerically(beta, "<", 180.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<", 180.0, eps=angle_tolerance)
        and compare_numerically(alpha, ">", 0.0, eps=angle_tolerance)
        and compare_numerically(beta, ">", 0.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">", 0.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<", alpha + beta, eps=angle_tolerance)
        and compare_numerically(alpha + beta, "<", 360.0 - gamma, eps=angle_tolerance)
        and compare_numerically(beta, "<", alpha + gamma, eps=angle_tolerance)
        and compare_numerically(alpha + gamma, "<", 360.0 - beta, eps=angle_tolerance)
        and compare_numerically(alpha, "<", beta + gamma, eps=angle_tolerance)
        and compare_numerically(beta + gamma, "<", 360.0 - alpha, eps=angle_tolerance)
    )

    if not result and raise_error:
        message = "Parameters could not form a parallelepiped:\n"
        message += f"a = {a}"
        if not compare_numerically(a, ">", 0.0, eps=length_tolerance):
            message += f" (a <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"b = {b}"
        if not compare_numerically(b, ">", 0.0, eps=length_tolerance):
            message += f" (b <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"c = {c}"
        if not compare_numerically(c, ">", 0.0, eps=length_tolerance):
            message += f" (c <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"alpha = {alpha}\n"
        if not compare_numerically(alpha, "<", 180.0, eps=angle_tolerance):
            message += f"  (alpha >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_numerically(alpha, ">", 0.0, eps=angle_tolerance):
            message += f"  (alpha <= 0 with numerical tolerance: {angle_tolerance})\n"
        message += f"beta = {beta}\n"
        if not compare_numerically(beta, "<", 180.0, eps=angle_tolerance):
            message += f"  (beta >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_numerically(beta, ">", 0.0, eps=angle_tolerance):
            message += f"  (beta <= 0 with numerical tolerance: {angle_tolerance})\n"
        message += f"gamma = {gamma}\n"
        if not compare_numerically(gamma, "<", 180.0, eps=angle_tolerance):
            message += f"  (gamma >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_numerically(gamma, ">", 0.0, eps=angle_tolerance):
            message += f"  (gamma <= 0 with numerical tolerance: {angle_tolerance})\n"
        if not compare_numerically(gamma, "<", alpha + beta, eps=angle_tolerance):
            message += f"Inequality gamma < alpha + beta is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_numerically(
            alpha + beta, "<", 360.0 - gamma, eps=angle_tolerance
        ):
            message += f"Inequality alpha + beta < 360 - gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_numerically(beta, "<", alpha + gamma, eps=angle_tolerance):
            message += f"Inequality beta < alpha + gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_numerically(
            alpha + gamma, "<", 360.0 - beta, eps=angle_tolerance
        ):
            message += f"Inequality alpha + gamma < 360 - beta is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_numerically(alpha, "<", beta + gamma, eps=angle_tolerance):
            message += f"Inequality alpha < beta + gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_numerically(
            beta + gamma, "<", 360.0 - alpha, eps=angle_tolerance
        ):
            message += f"Inequality beta + gamma < 360 - alpha is not satisfied with numerical tolerance: {angle_tolerance}\n"
        raise ValueError(message)

    return result


def absolute_to_relative(vector, basis):
    r"""
    Computes relative coordinates of the vector with respect to the basis.

    .. math::
        \boldsymbol{v} = v^1 \boldsymbol{e_1} + v^2 \boldsymbol{e_2} + v^3 \boldsymbol{e_3}

    First scalar products of the vector with the basis vectors are computed

    .. math::
        \begin{matrix}
        \boldsymbol{v} \cdot \boldsymbol{e_1}
        =
        v^1\, \boldsymbol{e_1} \cdot \boldsymbol{e_1}
        +
        v^2\, \boldsymbol{e_2} \cdot \boldsymbol{e_1}
        +
        v^3\, \boldsymbol{e_3} \cdot \boldsymbol{e_1} \\
        \boldsymbol{v} \cdot \boldsymbol{e_2}
        =
        v^1\, \boldsymbol{e_1} \cdot \boldsymbol{e_2}
        +
        v^2\, \boldsymbol{e_2} \cdot \boldsymbol{e_2}
        +
        v^3\, \boldsymbol{e_3} \cdot \boldsymbol{e_2} \\
        \boldsymbol{v} \cdot \boldsymbol{e_3}
        =
        v^1\, \boldsymbol{e_1} \cdot \boldsymbol{e_3}
        +
        v^2\, \boldsymbol{e_2} \cdot \boldsymbol{e_3}
        +
        v^3\, \boldsymbol{e_3} \cdot \boldsymbol{e_3}
        \end{matrix}

    Then, system of linear equations for :math:`v^1`, :math:`v^2`, :math:`v^3` is solved.

    Parameters
    ----------
    vector : (3,) |array-like|_
        Absolute coordinates of a vector.
    basis : (3, 3) |array-like|_
        Basis vectors. Rows are interpreted as vectors.
        Columns are interpreted as coordinates.

    Returns
    -------
    relative : (3,) :numpy:`ndarray`
        Relative coordinates of the ``vector`` with respect to the ``basis``.
        :math:`(v^1, v^2, v^3)`.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.geometry.absolute_to_relative([1, 0, 0], [[0, 1, 1], [1, 0, 1], [1, 1, 0]])
        array([-0.5,  0.5,  0.5])
        >>> wulf.geometry.absolute_to_relative([1, 0, 2.3241], [[0, 1, 1], [1, 0, 1], [1, 1, 0]])
        array([ 0.66205,  1.66205, -0.66205])

    """

    # Three vectors of the basis
    e1 = np.array(basis[0], dtype=float)
    e2 = np.array(basis[1], dtype=float)
    e3 = np.array(basis[2], dtype=float)

    v = np.array(vector, dtype=float)
    if (v == np.zeros(3)).all():
        return np.zeros(3)

    # Compose system of linear equations
    B = np.array([np.dot(e1, v), np.dot(e2, v), np.dot(e3, v)])
    A = np.array(
        [
            [np.dot(e1, e1), np.dot(e1, e2), np.dot(e1, e3)],
            [np.dot(e2, e1), np.dot(e2, e2), np.dot(e2, e3)],
            [np.dot(e3, e1), np.dot(e3, e2), np.dot(e3, e3)],
        ]
    )

    # Solve and return
    return np.linalg.solve(A, B)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from random import choices
from string import ascii_lowercase
from typing import Iterable

import numpy as np

from wulfric._kpoints_class import Kpoints
from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.cell._sc_standardize import get_conventional
from wulfric.cell._voronoi import _get_voronoi_cell
from wulfric.geometry._geometry import get_volume
from wulfric.visualization._interface import AbstractBackend

try:
    import matplotlib.pyplot as plt
    from matplotlib import rcParams
    from matplotlib.patches import FancyArrowPatch
    from mpl_toolkits.mplot3d import Axes3D, proj3d

    MATPLOTLIB_AVAILABLE = True
except ImportError:
    MATPLOTLIB_AVAILABLE = False


if MATPLOTLIB_AVAILABLE:
    # Better 3D arrows, see: https://stackoverflow.com/questions/22867620/putting-arrowheads-on-vectors-in-a-3d-plot
    class Arrow3D(FancyArrowPatch):
        def __init__(self, ax, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs
            self.ax = ax

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.ax.axes.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

        def do_3d_projection(self, *_, **__):
            return 0


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class MatplotlibBackend(AbstractBackend):
    r"""
    Plotting engine based on |matplotlib|_.

    Parameters
    ----------
    fig : matplotlib figure, optional
        Figure to plot on. If not provided, a new figure and ``ax`` is created.
    ax : matplotlib axis, optional
        Axis to plot on. If not provided, a new axis is created.
    background : bool, default True
        Whether to keep the axis in the plot.
    focal_length : float, default 0.2
        See: |matplotlibFocalLength|_

    Attributes
    ----------
    fig : matplotlib figure
        Figure to plot on.
    ax : matplotlib axis
        Axis to plot on.
    artists : dict
        Dictionary of the artists. Keys are the plot kinds, values are the lists of artists.

    Notes
    -----
    This class is a part of ``wulfric[visual]``
    """

    def __init__(self, fig=None, ax=None, background=True, focal_length=0.2):
        if not MATPLOTLIB_AVAILABLE:
            raise ImportError(
                'Matplotlib is not available. Install it with "pip install matplotlib"'
            )
        super().__init__()
        if fig is None:
            fig = plt.figure(figsize=(6, 6))
            ax = fig.add_subplot(projection="3d")
        elif ax is None:
            ax = fig.add_subplot(projection="3d")

        rcParams["axes.linewidth"] = 0
        rcParams["xtick.color"] = "#B3B3B3"
        ax.set_proj_type("persp", focal_length=focal_length)
        if background:
            ax.axes.linewidth = 0
            ax.xaxis._axinfo["grid"]["color"] = (1, 1, 1, 1)
            ax.yaxis._axinfo["grid"]["color"] = (1, 1, 1, 1)
            ax.zaxis._axinfo["grid"]["color"] = (1, 1, 1, 1)
            ax.set_xlabel("x", fontsize=15, alpha=0.5)
            ax.set_ylabel("y", fontsize=15, alpha=0.5)
            ax.set_zlabel("z", fontsize=15, alpha=0.5)
            ax.tick_params(axis="both", zorder=0, color="#B3B3B3")
        else:
            ax.axis("off")
        self.fig = fig
        self.ax = ax
        self.artists = {}

    def remove(self, kind="primitive"):
        r"""
        Removes a set of artists from the plot.

        Parameters
        ----------
        kind : str or list of str
            Type of the plot to be removed. Supported kinds:

            * "conventional"
            * "primitive"
            * "brillouin"
            * "kpath"
            * "brillouin_kpath"
            * "wigner_seitz"
        """

        if kind == "brillouin_kpath":
            kinds = ["brillouin", "kpath"]
        else:
            kinds = [kind]

        for kind in kinds:
            if kind not in self.artists:
                raise ValueError(f"No artists for the {kind} kind.")
            for artist in self.artists[kind]:
                if isinstance(artist, list):
                    for i in artist:
                        i.remove()
                else:
                    artist.remove()
            del self.artists[kind]
        self.ax.relim(visible_only=True)
        self.ax.set_aspect("equal")

    def plot(self, cell, kind="primitive", **kwargs):
        r"""
        Main plotting method.

        Actual list of supported kinds can be check with:

        .. doctest::

            >>> self.kinds.keys() # doctest: +SKIP

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        kind : str or list od str
            Type of the plot to be plotted. Supported plots:

            * "conventional"
            * "primitive"
            * "brillouin"
            * "kpath"
            * "brillouin-kpath"
            * "wigner-seitz"
            * "unit-cell"

        **kwargs
            Parameters to be passed to the specialized plotting function.
            See each function for the list of supported parameters.

        Raises
        ------
        ValueError
            If the plot kind is not supported.

        See Also
        --------
        plot_conventional : "conventional" plot.
        plot_primitive : "primitive" plot.
        plot_brillouin : "brillouin" plot.
        plot_kpath : "kpath" plot.
        plot_brillouin_kpath : "brillouin_kpath" plot.
        plot_wigner_seitz : "wigner-seitz" plot.
        plot_unit_cell : "unit-cell" plot.
        show : Shows the plot.
        save : Save the figure in the file.
        """

        super().plot(cell, kind=kind, **kwargs)
        self.ax.relim()
        self.ax.set_aspect("equal")

    def show(self, elev=30, azim=-60):
        r"""
        Shows the figure in the interactive mode.

        Parameters
        ----------
        elev : float, default 30
            Passed directly to matplotlib. See |matplotlibViewInit|_.
        azim : float, default -60
            Passed directly to matplotlib. See |matplotlibViewInit|_.
        """
        self.ax.set_aspect("equal")
        self.ax.view_init(elev=elev, azim=azim)
        plt.show()
        self.fig = None
        self.ax = None
        plt.close()

    def save(self, output_name="cell_graph.png", elev=30, azim=-60, **kwargs):
        r"""
        Saves the figure in the file.

        Parameters
        ----------
        output_name : str, default "cell_graph.png"
            Name of the file to be saved. With extension.
        elev : float, default 30
            Passed directly to matplotlib. See |matplotlibViewInit|_.
        azim : float, default -60
            Passed directly to matplotlib. See |matplotlibViewInit|_.
        **kwargs
            Parameters to be passed to the |matplotlibSavefig|_.
        """

        self.ax.set_aspect("equal")
        self.ax.view_init(elev=elev, azim=azim)
        self.fig.savefig(output_name, **kwargs)

    def clear(self):
        r"""
        Clears the axis.
        """

        if self.ax is not None:
            self.ax.cla()

    def legend(self, **kwargs):
        r"""
        Adds legend to the figure.

        Parameters
        ----------
        **kwargs :
            Directly passed to the |matplotlibLegend|_.
        """

        self.ax.legend(**kwargs)

    def plot_unit_cell(
        self,
        cell,
        vectors=True,
        color="#274DD1",
        label=None,
        vector_pad=1.1,
        conventional=False,
        reciprocal=False,
        normalize=False,
    ):
        r"""
        Plots real or reciprocal unit cell.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        vectors : bool, default True
            Whether to plot lattice vectors.
        color : str, default "#274DD1"
            Colour for the plot. Any format supported by matplotlib (see |matplotlibColor|_).
        label : str, optional
            Label for the plot.
        vector_pad : float, default 1.1
            Multiplier for the position of the vectors labels. 1 = position of the vector.
        conventional : bool, default False
            Whether to plot conventional cell.
            Only primitive unit cell is supported for reciprocal space.
        reciprocal : bool, default False
            Whether to plot reciprocal or real unit cell.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """
        if reciprocal and conventional:
            raise ValueError("Conventional cell is not supported in reciprocal space.")
        if conventional:
            artist_group = "conventional"
        else:
            artist_group = "primitive"

        if reciprocal:
            artist_group += "_reciprocal"
            vector_label = "b"
        else:
            artist_group += "_real"
            vector_label = "a"

        self.artists[artist_group] = []

        if conventional:
            cell = get_conventional(cell)
        elif reciprocal:
            cell = get_reciprocal(cell)

        if normalize:
            cell /= abs(get_volume(cell) ** (1 / 3.0))

        if label is not None:
            self.artists[artist_group].append(
                self.ax.scatter(0, 0, 0, color=color, label=label)
            )
        if vectors:
            if not isinstance(vector_pad, Iterable):
                vector_pad = [vector_pad, vector_pad, vector_pad]
            for i in range(3):
                self.artists[artist_group].append(
                    self.ax.text(
                        cell[i][0] * vector_pad[i],
                        cell[i][1] * vector_pad[i],
                        cell[i][2] * vector_pad[i],
                        f"${vector_label}_{i+1}$",
                        fontsize=20,
                        color=color,
                        ha="center",
                        va="center",
                    )
                )
                # Try beautiful arrows
                try:
                    self.artists[artist_group].append(
                        self.ax.add_artist(
                            Arrow3D(
                                self.ax,
                                [0, cell[i][0]],
                                [0, cell[i][1]],
                                [0, cell[i][2]],
                                mutation_scale=20,
                                arrowstyle="-|>",
                                color=color,
                                lw=2,
                                alpha=0.7,
                            )
                        )
                    )
                # Go to default
                except:
                    self.artists[artist_group].append(
                        self.ax.quiver(
                            0,
                            0,
                            0,
                            *tuple(cell[i]),
                            arrow_length_ratio=0.2,
                            color=color,
                            alpha=0.7,
                            linewidth=2,
                        )
                    )
                # Ghost point to account for the plot range
                self.artists[artist_group].append(self.ax.scatter(*tuple(cell[i]), s=0))

        def plot_line(line, shift):
            self.artists[artist_group].append(
                self.ax.plot(
                    [shift[0], shift[0] + line[0]],
                    [shift[1], shift[1] + line[1]],
                    [shift[2], shift[2] + line[2]],
                    color=color,
                )
            )

        for i in range(0, 3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            plot_line(cell[i], np.zeros(3))
            plot_line(cell[i], cell[j])
            plot_line(cell[i], cell[k])
            plot_line(cell[i], cell[j] + cell[k])

    def plot_wigner_seitz(
        self,
        cell,
        vectors=True,
        color="black",
        label=None,
        vector_pad=1.1,
        reciprocal=False,
        normalize=False,
    ):
        r"""
        Plots Wigner-Seitz cell.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        vectors : bool, default True
            Whether to plot lattice vectors.
        color : str, default "black"
            Colour for the plot. Any format supported by matplotlib (see |matplotlibColor|_).
        label : str, optional
            Label for the plot.
        vector_pad : float, default 1.1
            Multiplier for the position of the vectors labels. 1 = position of the vector.
        reciprocal : bool, default False
            Whether to plot reciprocal or real Wigner-Seitz cell.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """

        if reciprocal:
            artist_group = "brillouin"
        else:
            artist_group = "wigner_seitz"

        self.artists[artist_group] = []

        if reciprocal:
            cell = get_reciprocal(cell)
            vector_label = "b"
        else:
            vector_label = "a"

        if color is None:
            color = "black"

        if normalize:
            cell /= abs(get_volume(cell) ** (1 / 3.0))

        v1, v2, v3 = cell[0], cell[1], cell[2]

        vs = [v1, v2, v3]

        if label is not None:
            self.artists[artist_group].append(
                self.ax.scatter(0, 0, 0, color=color, label=label)
            )
        if vectors:
            if not isinstance(vector_pad, Iterable):
                vector_pad = [vector_pad, vector_pad, vector_pad]
            for i in range(3):
                self.artists[artist_group].append(
                    self.ax.text(
                        vs[i][0] * vector_pad[i],
                        vs[i][1] * vector_pad[i],
                        vs[i][2] * vector_pad[i],
                        f"${vector_label}_{i+1}$",
                        fontsize=20,
                        color=color,
                        ha="center",
                        va="center",
                    )
                )
                # Try beautiful arrows
                try:
                    self.artists[artist_group].append(
                        self.ax.add_artist(
                            Arrow3D(
                                self.ax,
                                [0, vs[i][0]],
                                [0, vs[i][1]],
                                [0, vs[i][2]],
                                mutation_scale=20,
                                arrowstyle="-|>",
                                color=color,
                                lw=2,
                                alpha=0.8,
                            )
                        )
                    )
                # Go to default
                except:
                    self.artists[artist_group].append(
                        self.ax.quiver(
                            0,
                            0,
                            0,
                            *tuple(vs[i]),
                            arrow_length_ratio=0.2,
                            color=color,
                            alpha=0.5,
                        )
                    )
                # Ghost point to account for the plot range
                self.artists[artist_group].append(self.ax.scatter(*tuple(vs[i]), s=0))

        edges, _ = _get_voronoi_cell(cell)
        for p1, p2 in edges:
            self.artists[artist_group].append(
                self.ax.plot(
                    [p1[0], p2[0]],
                    [p1[1], p2[1]],
                    [p1[2], p2[2]],
                    color=color,
                )
            )

    def plot_kpath(self, cell, color="black", label=None, normalize=False):
        r"""
        Plots k path in the reciprocal space.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        color : str, default "black"
            Colour for the plot. Any format supported by matplotlib (see |matplotlibColor|_).
        label : str, optional
            Label for the plot.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """

        artist_group = "kpath"

        self.artists[artist_group] = []

        if normalize:
            cell /= get_volume(cell) ** (1 / 3.0)

        kp = Kpoints.from_cell(cell)

        cell = get_reciprocal(cell)

        for point in kp.hs_names:
            self.artists[artist_group].append(
                self.ax.scatter(
                    *tuple(kp.hs_coordinates[point] @ cell),
                    s=36,
                    color=color,
                )
            )

            self.artists[artist_group].append(
                self.ax.text(
                    *tuple(
                        kp.hs_coordinates[point] @ cell
                        + 0.025 * cell[0]
                        + +0.025 * cell[1]
                        + 0.025 * cell[2]
                    ),
                    kp.hs_labels[point],
                    fontsize=20,
                    color=color,
                )
            )
        if label is not None:
            self.artists[artist_group].append(
                self.ax.scatter(
                    0,
                    0,
                    0,
                    s=36,
                    color=color,
                    label=label,
                )
            )

        for subpath in kp.path:
            for i in range(len(subpath) - 1):
                self.artists[artist_group].append(
                    self.ax.plot(
                        *tuple(
                            np.concatenate(
                                (
                                    kp.hs_coordinates[subpath[i]] @ cell,
                                    kp.hs_coordinates[subpath[i + 1]] @ cell,
                                )
                            )
                            .reshape(2, 3)
                            .T
                        ),
                        color=color,
                        alpha=0.5,
                        linewidth=3,
                    )
                )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import acos

import numpy as np

from wulfric._exceptions import NiggliReductionFailed
from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.cell._niggli import niggli
from wulfric.constants._numerical import TODEGREES

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


################################################################################
#                                  LePage CUB                                  #
################################################################################
def _check_cub(angles, axes, angle_tolerance):
    target_angles = np.array(
        [
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
        ]
    )

    conventional_axis = np.array([0, 45, 45, 45, 45, 90, 90, 90, 90])

    axes = np.array([i[0] for i in axes])

    if 9 <= angles.shape[0]:
        sub_angles = angles[:9, :9]
        sub_axes = axes[:9]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < angle_tolerance
        ).all():
            xyz = []
            for i in range(9):
                if (
                    np.abs(np.sort(sub_angles[i]) - conventional_axis) < angle_tolerance
                ).all():
                    xyz.append(sub_axes[i])
            det = np.abs(np.linalg.det(xyz))
            if det == 1:
                result = "CUB"
            elif det == 4:
                result = "FCC"
            elif det == 2:
                result = "BCC"
            return result, False
        return None, True
    return None, True


################################################################################
#                                  LePage HEX                                  #
################################################################################
def _check_hex(angles, angle_tolerance):
    target_angles = np.array(
        [
            [0, 90, 90, 90, 90, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
        ]
    )
    if 7 <= angles.shape[0]:
        sub_angles = angles[:7, :7]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < angle_tolerance
        ).all():
            return "HEX", False
        return None, True
    return None, True


################################################################################
#                                  LePage TET                                  #
################################################################################
def _check_tet(angles, axes, angle_tolerance, cell):
    target_angles = np.array(
        [
            [0, 90, 90, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
        ]
    )

    conventional_axis = np.array([0, 90, 90, 90, 90])

    axes = np.array([i[0] for i in axes])

    if 5 <= angles.shape[0]:
        sub_angles = angles[:5, :5]
        sub_axes = axes[:5]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < angle_tolerance
        ).all():
            xy = []
            for i in range(5):
                if (
                    np.abs(np.sort(sub_angles[i]) - conventional_axis) < angle_tolerance
                ).all():
                    z = sub_axes[i]
                else:
                    xy.append(sub_axes[i])
            xy.sort(key=lambda x: np.linalg.norm(x @ cell))

            xyz = [xy[0], xy[1], z]

            det = np.abs(np.linalg.det(xyz))
            if det == 1:
                result = "TET"
            elif det == 2:
                result = "BCT"
            return result, False
        return None, True
    return None, True


################################################################################
#                                  LePage RHL                                  #
################################################################################
def _check_rhl(angles, angle_tolerance):
    target_angles = np.array(
        [
            [0, 60, 60],
            [0, 60, 60],
            [0, 60, 60],
        ]
    )

    if 3 <= angles.shape[0]:
        sub_angles = angles[:3, :3]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < angle_tolerance
        ).all():
            return "RHL", False
        return None, True
    return None, True


################################################################################
#                                  LePage ORC                                  #
################################################################################
def _check_orc(angles, axes, angle_tolerance):
    target_angles = np.array(
        [
            [0, 90, 90],
            [0, 90, 90],
            [0, 90, 90],
        ]
    )

    axes = np.array([i[0] for i in axes])
    if 3 <= angles.shape[0]:
        sub_angles = angles[:3, :3]
        sub_axes = axes[:3]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < angle_tolerance
        ).all():
            C = np.array(sub_axes, dtype=float).T
            det = np.abs(np.linalg.det(C))
            if det == 1:
                result = "ORC"
            if det == 4:
                result = "ORCF"
            if det == 2:
                v = C @ [1, 1, 1]

                def gcd(p, q):
                    while q != 0:
                        p, q = q, p % q
                    return p

                if (
                    gcd(abs(v[0]), abs(v[1])) > 1
                    and gcd(abs(v[0]), abs(v[2])) > 1
                    and gcd(abs(v[1]), abs(v[2])) > 1
                ):
                    result = "ORCI"
                else:
                    result = "ORCC"
            return result, False
        return None, True
    return None, True


################################################################################
#                                  LePage MCL                                  #
################################################################################
def _get_perpendicular_shortest(v, cell, angle_tolerance):
    perp_axes = []

    miller_indices = (np.indices((3, 3, 3)) - 1).transpose((1, 2, 3, 0)).reshape(27, 3)

    for index in miller_indices:
        if (index != [0, 0, 0]).any():
            if abs((index @ cell) @ (v @ cell)) < angle_tolerance:
                perp_axes.append(index)

    perp_axes.sort(key=lambda x: np.linalg.norm(x @ cell))

    # indices 0 and 2 (not 0 and 1), since v and -v are present in miller_indices
    return perp_axes[0], perp_axes[2]


def _check_mcl(angles, axes, angle_tolerance, cell):
    axes = np.array([i[0] for i in axes])
    angles = angles[:1]
    if 1 <= angles.shape[0]:
        b = axes[0]

        # If we are here by mistake it can fail
        try:
            a, c = _get_perpendicular_shortest(b, cell, angle_tolerance)
        except IndexError:
            return None, True
        C = np.array([a, b, c], dtype=float).T
        det = np.abs(np.linalg.det(C))
        if det == 1:
            return "MCL", False
        if det == 2:
            return "MCLC", False
        return None, True
    return None, True


################################################################################
#                                    LePage                                    #
################################################################################
def lepage(
    cell, angle_tolerance=1e-4, give_all_results=False, no_niggli=False, _limit=3.0
):
    r"""
    Detect Bravais lattice type with the Le Page algorithm [1]_.

    Details of the implementation are written in :ref:`library_lepage`.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    angle_tolerance : float, default 1e-4
        Angle tolerance for the search of the actual symmetry axes. It is recommended to
        reduce ``angle_tolerance`` to account for the finite precision of the angles of
        the ``cell``. Default value is chosen in the contexts of condense matter physics,
        assuming that angles are in degrees. Please choose appropriate tolerance for your
        problem.
    give_all_results : bool, default False
        Whether to return the list of Bravais lattice types identified during the
        process of exclusion of the pseudosymmetry axes. Last element is the computed
        Bravais lattice type.
    no_niggli : bool, default False
        Whether to skip niggli reduction.
    _limit : float, default 3.0
        Tolerance parameter for the construction of the list of potential symmetry axes.
        Given in degrees. Change with caution and only if you understand what this
        parameter means and read [1]_.

    Returns
    -------
    lattice_type : str
        Bravais lattice type.
    intermediate_types : list of str
        Returned if ``give_all_results`` is ``True``.

    References
    ----------
    .. [1] Le Page, Y., 1982.
        The derivation of the axes of the conventional unit cell from
        the dimensions of the Buerger-reduced cell.
        Journal of Applied Crystallography, 15(3), pp.255-259.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.lepage([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        'CUB'
        >>> wulf.cell.lepage([[1, 0, 0], [0, 1, 0], [0, 0, 2]])
        'TET'
        >>> wulf.cell.lepage([[1, 0, 0], [0, 2, 0], [0, 0, 3]])
        'ORC'
    """

    # Safeguard to avoid infinite loops
    angle_tolerance = abs(angle_tolerance)

    if not no_niggli:
        # Niggli reduction
        try:
            cell = niggli(cell=cell)
        except NiggliReductionFailed:
            import warnings

            warnings.warn(
                "LePage algorithm: Niggli reduction failed, using input cell",
                RuntimeWarning,
            )

    rcell = get_reciprocal(cell)

    # Find all potential axes, including twins
    miller_indices = (np.indices((5, 5, 5)) - 2).transpose((1, 2, 3, 0)).reshape(125, 3)
    axes = []
    for U in miller_indices:
        for h in miller_indices:
            if abs(U @ h) == 2:
                t = U @ cell
                tau = h @ rcell
                delta = (
                    np.arctan(np.linalg.norm(np.cross(t, tau)) / abs(t @ tau))
                    * TODEGREES
                )
                if delta < _limit:
                    axes.append([U, t / np.linalg.norm(t), abs(U @ h), delta])

    # Sort and filter
    axes.sort(key=lambda x: x[-1])
    keep_index = np.ones(len(axes))
    for i in range(len(axes)):
        if keep_index[i]:
            for j in range(i + 1, len(axes)):
                if (
                    (axes[i][0] == axes[j][0]).all()
                    or (axes[i][0] == -axes[j][0]).all()
                    or (axes[i][0] == 2 * axes[j][0]).all()
                    or (axes[i][0] == 0.5 * axes[j][0]).all()
                ):
                    keep_index[i] = 0
                    break
    new_axes = []
    for i in range(len(axes)):
        if keep_index[i]:
            if set(axes[i][0]) == {0, 2}:
                axes[i][0] = axes[i][0] / 2
            new_axes.append(axes[i])
    axes = new_axes

    # Compute matrix of pair-wise angles
    n = len(axes)
    angles = np.zeros((n, n), dtype=float)
    for i in range(n):
        for j in range(i, n):
            angles[i][j] = (
                acos(abs(np.clip(np.array(axes[i][1]) @ np.array(axes[j][1]), -1, 1)))
                * TODEGREES
            )
    angles += angles.T

    # Main cycle
    delta = None

    if give_all_results:
        results = []

    while delta is None or delta > angle_tolerance:

        if len(axes) == 0:
            delta = 0
        else:
            delta = max(axes, key=lambda x: x[-1])[-1]

        continue_search = True
        n = len(axes)
        result = None

        # CUB
        result, continue_search = _check_cub(angles, axes, angle_tolerance)

        # HEX
        if continue_search:
            result, continue_search = _check_hex(angles, angle_tolerance)

        # TET
        if continue_search:
            result, continue_search = _check_tet(angles, axes, angle_tolerance, cell)

        # RHL
        if continue_search:
            result, continue_search = _check_rhl(angles, angle_tolerance)

        # ORC
        if continue_search:
            result, continue_search = _check_orc(angles, axes, angle_tolerance)

        # MCL
        if continue_search:
            result, continue_search = _check_mcl(angles, axes, angle_tolerance, cell)

        # TRI
        if continue_search:
            result = "TRI"

        if len(axes) > 0:
            # remove worst axes
            while len(axes) >= 2 and axes[-1][-1] == axes[-2][-1]:
                axes = axes[:-1]
                angles = angles[:-1, :-1]
            axes = axes[:-1]
            angles = angles[:-1, :-1]

        if give_all_results:
            results.append(result)

    if give_all_results:
        return results

    return result


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from wulfric.cell._lepage import lepage
from wulfric.cell._sc_standardize import get_S_matrix, get_standardized

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def standardize(
    cell, atoms, S_matrix=None, length_tolerance=1e-8, angle_tolerance=1e-4
):
    R"""
    Standardize cell with respect to the Bravais lattice type as defined in [1]_ and
    update atom's relative coordinates.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.
    S_matrix : (3, 3) |array-like|_, optional
        Transformation matrix S. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors).  Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Standardized cell. Rows are interpreted as vectors. Independent from the initial
        cell, safe to modify.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    Notes
    -----

    ``atoms`` are not returned, but rather updated.

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import numpy as np
        >>> import wulfric as wulf
        >>> cell = np.array([[2, 0, 0], [0, 1, 0], [0, 0, 3]])
        >>> atoms = {"names" : ["Cr1", "Cr2"], "positions" : [[0.7, 0, 0], [0, 0.2, 0]]}
        >>> atoms["positions"][0] @ cell, atoms["positions"][1] @ cell
        (array([1.4, 0. , 0. ]), array([0. , 0.2, 0. ]))
        >>> cell = wulf.crystal.standardize(cell, atoms)
        >>> cell
        array([[ 0., -1.,  0.],
               [-2.,  0.,  0.],
               [ 0.,  0., -3.]])
        >>> atoms
        {'names': ['Cr1', 'Cr2'], 'positions': [array([ 0. , -0.7,  0. ]), array([-0.2,  0. ,  0. ])]}
        >>> # Note that absolute coordinates of atoms are not changed.
        >>> atoms["positions"][0] @ cell, atoms["positions"][1] @ cell
        (array([1.4, 0. , 0. ]), array([0. , 0.2, 0. ]))

    """

    cell = np.array(cell, dtype=float)

    # Get S matrix before cell standardization
    if S_matrix is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

        S_matrix = get_S_matrix(
            cell,
            lattice_type,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )
    else:
        S_matrix = np.array(S_matrix, dtype=float)

    # Standardize cell
    cell = get_standardized(cell=cell, S_matrix=S_matrix)

    # Recalculate atom's relative coordinates.
    atoms["positions"] = [
        np.linalg.inv(S_matrix) @ position for position in atoms["positions"]
    ]

    return cell


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from copy import deepcopy
from typing import Iterable

import numpy as np

from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.cell._kpoints import get_hs_data
from wulfric.cell._lepage import lepage
from wulfric.geometry._geometry import absolute_to_relative

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class Kpoints:
    r"""
    Interface for convenient manipulation with the high symmetry kpoints and K-path in
    reciprocal space.

    Parameters
    ----------
    rcell : (3, 3) |array-like|_
        Reciprocal cell. Rows are interpreted as vectors.
    coordinates : list, optional
        Coordinates of high symmetry points given in relative coordinates in reciprocal
        space.
    names: list, optional
        Names of the high symmetry points. Used in ``path``. Has to have the same length
        as ``coordinates``. If ``None``, then use "K1", ... "KN", where
        ``N = len(coordinates)``.
    labels : list, optional
        List of the high symmetry point's labels. Used for plotting. Has to have the same
        length as ``coordinates``. If ``None``, then use "K$_1$", ... "K$_N$", where
        ``N = len(coordinates)``.
    path : str, optional
        K-path. Use elements of ``names`` to specify the path. If no names given, then use
        "K1", ... "KN", where ``N = len(coordinates)``.
    n : int
        Number of intermediate points between each pair of the high symmetry points (high
        symmetry points excluded).

    Attributes
    ----------
    rcell : (3, 3) :numpy:`ndarray`
        Reciprocal cell. Rows are interpreted as vectors.
    hs_names : list
        Names of the high symmetry points. Used for programming, not for plotting.
    hs_coordinates : dict
        Dictionary of the high symmetry points coordinates.

        .. code-block:: python

            {"name": [k_a, k_b, k_c], ... }

    hs_labels : dict
        Dictionary of the high symmetry points labels for plotting.

        .. code-block:: python

            {"name": "label", ... }
    """

    def __init__(
        self, rcell, coordinates=None, names=None, labels=None, path=None, n=100
    ) -> None:
        self.rcell = np.array(rcell)

        if coordinates is None:
            coordinates = []

        # Fill names and labels with defaults
        if names is None:
            names = [f"K{i+1}" for i in range(len(coordinates))]
            if labels is None:
                labels = [f"K$_{i+1}$" for i in range(len(coordinates))]
        if labels is None:
            labels = [name for name in names]
        else:
            if len(labels) != len(coordinates):
                raise ValueError(
                    f"Amount of labels ({len(labels)}) does not match amount of points ({len(coordinates)})."
                )

        # Define high symmetry points attributes
        self.hs_coordinates = dict(
            [(names[i], np.array(coordinates[i])) for i in range(len(coordinates))]
        )
        self.hs_labels = dict([(names[i], labels[i]) for i in range(len(coordinates))])
        self.hs_names = names

        self._n = n

        self._path = None
        if path is None:
            path = "-".join(self.hs_names)
        self.path = path

    @staticmethod
    def from_cell(
        cell,
        lattice_type=None,
        lattice_variation=None,
        S_matrix=None,
        C_matrix=None,
        length_tolerance=1e-8,
        angle_tolerance=1e-4,
        n=100,
    ):
        r"""
        Creates an instance of the :py:class:`wulf.Kpoints` from ``cell``.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        lattice_type : str, optional
            One of the 14 lattice types that correspond to the provided ``cell``,
            case-insensitive. If not provided, then computed automatically from ``cell``.
            If provided, then it user's responsibility to ensure that ``lattice_type`` is
            correct.
        lattice_variation : str, optional
            One of the lattice variations that correspond to the provided ``cell`` and
            ``lattice_type``. If not provided, then computed automatically. Case-insensitive.
        S_matrix : (3, 3) |array-like|_, optional
            Transformation matrix S. If not provided, then computed automatically from
            ``cell``. If provided, then it is user's responsibility to ensure that the matrix
            is the correct one for the given ``cell``.
        C_matrix : (3, 3) |array-like|_, optional
            Transformation matrix C. If not provided, then computed automatically from
            ``cell``. If provided, then it is user's responsibility to ensure that the matrix
            is the correct one for the given ``cell``.
        length_tolerance : float, default :math:`10^{-8}`
            Tolerance for length variables (lengths of the lattice vectors). Default
            value is chosen in the contexts of condense matter physics, assuming that
            length is given in Angstroms. Please choose appropriate tolerance for your
            problem.
        angle_tolerance : float, default :math:`10^{-4}`
            Tolerance for angle variables (angles of the lattice). Default value is chosen
            in the contexts of condense matter physics, assuming that angles are in
            degrees. Please choose appropriate tolerance for your problem.
        n : int, default 100
            Number of points between each pair of the high symmetry points
            (high symmetry points excluded).

        Returns
        -------
        kp : :py:class:`wulf.Kpoints`
        """

        coordinates, names, labels, path = get_hs_data(
            cell,
            lattice_type=lattice_type,
            lattice_variation=lattice_variation,
            S_matrix=S_matrix,
            C_matrix=C_matrix,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )

        return Kpoints(
            get_reciprocal(cell),
            coordinates=coordinates,
            names=names,
            labels=labels,
            path=path,
            n=n,
        )

    ################################################################################
    #                            High symmetry points                              #
    ################################################################################
    def add_hs_point(self, name, coordinate, label, relative=True) -> None:
        r"""
        Adds high symmetry point.

        Parameters
        ----------
        name : str
            Name of the high symmetry point.
        coordinate : (3,) array-like
            Coordinate of the high symmetry point.
        label : str
            Label of the high symmetry point, ready to be plotted.
        relative : bool, optional
            Whether to interpret coordinates as relative or absolute.
        """

        if name in self.hs_names:
            raise ValueError(f"Point '{name}' already defined.")

        if not relative:
            coordinate = absolute_to_relative(coordinate, self.rcell)

        self.hs_names.append(name)
        self.hs_coordinates[name] = np.array(coordinate)
        self.hs_labels[name] = label

    def remove_hs_point(self, name) -> None:
        r"""
        Removes high symmetry point.

        Parameters
        ----------
        name : str
            Name of the high symmetry point.
        """

        if name in self.hs_names:
            self.hs_names.remove(name)
            del self.hs_coordinates[name]
            del self.hs_labels[name]

    ################################################################################
    #                                Path attributes                               #
    ################################################################################
    @property
    def path(self) -> list:
        r"""
        K points path.

        Returns
        -------
        path : list of list of str
            K points path. Each subpath is a list of the high symmetry points.
        """

        return self._path

    @path.setter
    def path(self, new_path):
        if isinstance(new_path, str):
            tmp_path = new_path.split("|")
            new_path = []
            for i in range(len(tmp_path)):
                subpath = tmp_path[i].split("-")
                # Each subpath has to contain at least two points.
                if len(subpath) != 1:
                    new_path.append(subpath)
        elif isinstance(new_path, Iterable):
            tmp_path = new_path
            new_path = []
            for subpath in tmp_path:
                if isinstance(subpath, str) and "-" in subpath:
                    subpath = subpath.split("-")
                    # Each subpath has to contain at least two points.
                    if len(subpath) != 1:
                        new_path.append(subpath)
                elif (
                    not isinstance(subpath, str)
                    and isinstance(subpath, Iterable)
                    and len(subpath) != 1
                ):
                    new_path.append(subpath)
                else:
                    new_path = [tmp_path]
                    break
        # Check if all points are defined.
        for subpath in new_path:
            for point in subpath:
                if point not in self.hs_names:
                    message = f"Point '{point}' is not defined. Defined points are:"
                    for defined_name in self.hs_names:
                        message += (
                            f"\n  {defined_name} : {self.hs_coordinates[defined_name]}"
                        )
                    raise ValueError(message)
        self._path = new_path

    @property
    def path_string(self) -> str:
        r"""
        K points path as a string.

        Returns
        -------
        path : str
        """

        result = ""
        for s_i, subpath in enumerate(self.path):
            for i, name in enumerate(subpath):
                if i != 0:
                    result += "-"
                result += name
            if s_i != len(self.path) - 1:
                result += "|"

        return result

    @property
    def n(self) -> int:
        r"""
        Amount of points between each pair of the high symmetry points
        (high symmetry points excluded).

        Returns
        -------
        n : int
        """

        return self._n

    @n.setter
    def n(self, new_n):
        if not isinstance(new_n, int):
            raise ValueError(
                f"n has to be integer. Given: {new_n}, type = {type(new_n)}"
            )
        self._n = new_n

    ################################################################################
    #                         Attributes for the axis ticks                        #
    ################################################################################
    @property
    def labels(self) -> list:
        r"""
        Labels of high symmetry points, ready to be plotted.

        For example for point "Gamma" it returns r"$\Gamma$".

        If there are two high symmetry points following one another in the path,
        it returns "X|Y" where X and Y are the labels of the two high symmetry points.

        Returns
        -------
        labels : list of str
            Labels, ready to be plotted. Same length as :py:attr:`.ticks`.
        """

        labels = []
        for s_i, subpath in enumerate(self.path):
            if s_i != 0:
                labels[-1] += "|" + self.hs_labels[subpath[0]]
            else:
                labels.append(self.hs_labels[subpath[0]])
            for name in subpath[1:]:
                labels.append(self.hs_labels[name])

        return labels

    def ticks(self, relative=False):
        r"""
        Tick's positions of the high symmetry points, ready to be plotted.

        Parameters
        ----------
        relative : bool, optional
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        ticks : :numpy:`ndarray`
            Tick's positions, ready to be plotted. Same length as :py:attr:`.labels`.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        ticks = []
        for s_i, subpath in enumerate(self.path):
            if s_i == 0:
                ticks.append(0)
            for i, name in enumerate(subpath[1:]):
                ticks.append(
                    np.linalg.norm(
                        self.hs_coordinates[name] @ cell
                        - self.hs_coordinates[subpath[i]] @ cell
                    )
                    + ticks[-1]
                )

        return np.array(ticks)

    ################################################################################
    #                   Points of the path with intermediate ones                  #
    ################################################################################
    def points(self, relative=False):
        r"""
        Coordinates of all points with n points between each pair of the high
        symmetry points (high symmetry points excluded).

        Parameters
        ----------
        relative : bool, optional
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        points : (N, 3) :numpy:`ndarray`
            Coordinates of all points.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        points = None
        for subpath in self.path:
            for i in range(len(subpath) - 1):
                name = subpath[i]
                next_name = subpath[i + 1]
                new_points = np.linspace(
                    self.hs_coordinates[name] @ cell,
                    self.hs_coordinates[next_name] @ cell,
                    self._n + 2,
                )
                if points is None:
                    points = new_points
                else:
                    points = np.concatenate((points, new_points))
        return points

    # It can not just call for points and flatten them, because it has to treat "|" as a special case.
    def flatten_points(self, relative=False):
        r"""
        Flatten coordinates of all points with n points between each pair of the high
        symmetry points (high symmetry points excluded).

        Used to plot band structure, dispersion, etc.

        Parameters
        ----------
        relative : bool, optional
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        flatten_points : (N, 3) :numpy:`ndarray`
            Flatten coordinates of all points.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        flatten_points = None
        for s_i, subpath in enumerate(self.path):
            for i in range(len(subpath) - 1):
                name = subpath[i]
                next_name = subpath[i + 1]
                points = (
                    np.linspace(
                        self.hs_coordinates[name] @ cell,
                        self.hs_coordinates[next_name] @ cell,
                        self._n + 2,
                    )
                    - self.hs_coordinates[name] @ cell
                )
                delta = np.linalg.norm(points, axis=1)
                if s_i == 0 and i == 0:
                    flatten_points = delta
                else:
                    delta += flatten_points[-1]
                    flatten_points = np.concatenate((flatten_points, delta))
        return flatten_points

    ################################################################################
    #                                     Copy                                     #
    ################################################################################

    def copy(self):
        r"""
        Creates a copy of the kpoints.

        Returns
        -------
        kpoints : :py:class:`.Kpoints`
            Copy of the kpoints.
        """

        return deepcopy(self)

    ################################################################################
    #                                Human readables                               #
    ################################################################################

    def hs_table(self, decimals=8) -> str:
        r"""
        Table of the high symmetry points.

        Parameters
        ----------
        decimals : int, optional
            Number of decimal places to round the coordinates.

        Returns
        -------
        table : str
            String with N+1 lines, where N is the amount of high symmetry points.
            Each line contains the name of the high symmetry point and its relative and
            absolute coordinates in a reciprocal space, i.e.::

                K1  0.0 0.0 0.0   0.0 0.0 0.0

            First line is a header::

                Name  rel_b1 rel_b2 rel_b3  k_x k_y k_z
        """

        d = decimals
        table = [
            (
                f"{'Name':4}  "
                + f"{'rel_b1':>{d+3}} "
                + f"{'rel_b2':>{d+3}} "
                + f"{'rel_b3':>{d+3}}  "
                + f"{'k_x':>{d+3}} "
                + f"{'k_y':>{d+3}} "
                + f"{'k_z':>{d+3}}"
            )
        ]
        for name in self.hs_names:
            relative = self.hs_coordinates[name]
            i = f"{relative[0]: {d+3}.{d}f}"
            j = f"{relative[1]: {d+3}.{d}f}"
            k = f"{relative[2]: {d+3}.{d}f}"
            absolute = self.hs_coordinates[name] @ self.rcell
            k_x = f"{absolute[0]: {d+3}.{d}f}"
            k_y = f"{absolute[1]: {d+3}.{d}f}"
            k_z = f"{absolute[2]: {d+3}.{d}f}"
            table.append(f"{name:4}  {i} {j} {k}  {k_x} {k_y} {k_z}")
        return "\n".join(table)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from calendar import month_name
from datetime import datetime

from wulfric import __doclink__, __release_date__, __version__

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def logo(info=None, line_length=None, flat=False, date_time=False, comment=None):
    """
    Generates logo of wulfric with some information about the package

    Parameters
    ----------
    info : list of str, optional
        Information about the package, that is displayed below the logo.
        Each element should not exceed 50 characters.
        by default it displays the version, release date,
        git hash and documentation link. You can pass th empty list
        to display only the logo.
    line_length : int, optional
        Length of the lines to be returned.
        Minimum value is 62.
    flat : bool
        Whether to return a flat logo or not.
    date_time : bool, default False
        Whether to include the date and time to the standard info or not.
    comment : str or bool, optional
        Whether to use some character at the end of each string. If bool and
        True, then "# " is used. If str, then this string is used. If None, then
        no character is used.

    Returns
    -------
    logo_info : str
        Logo and information about the package.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> print(wulf.logo(info=[])) # doctest: +NORMALIZE_WHITESPACE
        ██╗      ██╗ ██╗   ██╗ ██╗      ███████╗ ██████╗  ██╗  ██████╗
        ██║      ██║ ██║   ██║ ██║      ██╔════╝ ██╔══██╗ ██║ ██╔════╝
        ██║  █╗  ██║ ██║   ██║ ██║      █████╗   ██████╔╝ ██║ ██║
        ██║ ███╗ ██║ ██║   ██║ ██║      ██╔══╝   ██╔══██╗ ██║ ██║
        ╚███╔══███╔╝ ╚██████╔╝ ███████╗ ██║      ██║  ██║ ██║ ╚██████╗
        ╚══╝  ╚══╝   ╚═════╝  ╚══════╝ ╚═╝      ╚═╝  ╚═╝ ╚═╝  ╚═════╝
    """
    if info is None:
        info = [
            f"Version: {__version__}",
            f"Documentation: {__doclink__}",
            f"Release date: {__release_date__}",
            f"Licence: GNU GPLv3",
            f"Copyright (C) 2023-{datetime.now().year}  Andrey Rybakov",
        ]
        if date_time:
            cd = datetime.now()
            info.append("")
            info.append(
                f"Generated on {cd.day} {month_name[cd.month]} {cd.year}"
                + f" at {cd.hour}:{cd.minute}:{cd.second} "
            )
    logo = [
        "██╗      ██╗ ██╗   ██╗ ██╗      ███████╗ ██████╗  ██╗  ██████╗",
        "██║      ██║ ██║   ██║ ██║      ██╔════╝ ██╔══██╗ ██║ ██╔════╝",
        "██║  █╗  ██║ ██║   ██║ ██║      █████╗   ██████╔╝ ██║ ██║     ",
        "██║ ███╗ ██║ ██║   ██║ ██║      ██╔══╝   ██╔══██╗ ██║ ██║     ",
        "╚███╔══███╔╝ ╚██████╔╝ ███████╗ ██║      ██║  ██║ ██║ ╚██████╗",
        " ╚══╝  ╚══╝   ╚═════╝  ╚══════╝ ╚═╝      ╚═╝  ╚═╝ ╚═╝  ╚═════╝",
    ]
    if flat:
        logo = [
            "██       ██  ██    ██  ██       ███████  ██████   ██   ██████",
            "██       ██  ██    ██  ██       ██       ██   ██  ██  ██     ",
            "██   █   ██  ██    ██  ██       █████    ██████   ██  ██     ",
            "██  ███  ██  ██    ██  ██       ██       ██   ██  ██  ██     ",
            " ███   ███    ██████   ███████  ██       ██   ██  ██   ██████",
        ]
    cat = [
        "▄   ▄     ",
        "█▀█▀█     ",
        "█▄█▄█     ",
        " ███   ▄▄ ",
        " ████ █  █",
        " ████    █",
        " ▀▀▀▀▀▀▀▀ ",
    ]

    N = len(logo[0])
    n = len(cat[0]) + 2
    if line_length is None:
        line_length = N
    if line_length < N:
        line_length = N
    if isinstance(comment, bool) and comment:
        comment = "# "
    elif comment is not None:
        comment = str(comment)
    else:
        comment = ""

    logo_info = [f"{x:^{N}}" for x in logo]
    if len(info) > 0:
        if len(info) <= len(cat):
            before = (len(cat) - len(info)) // 2 + (len(cat) - len(info)) % 2
            after = len(cat) - len(info) - before
            for i in range(len(cat)):
                if i < before or i >= len(cat) - after:
                    logo_info.append(f"{' ':{N-n}}{cat[i]:^{n}}")
                else:
                    logo_info.append(f"{info[i-before]:^{N-n}}{cat[i]:^{n}}")
        else:
            before = (len(info) - len(cat)) // 2
            after = len(info) - len(cat) - before
            for i in range(len(info)):
                if i < before or i >= len(info) - after:
                    logo_info.append(f"{info[i-before]:^{N-n}}")
                else:
                    logo_info.append(f"{info[i-before]:^{N-n}}{cat[i-before]:^{n}}")

    logo_info = [f"{comment}{x:^{line_length}}\n" for x in logo_info]
    return "".join(logo_info)[:-1]


def _copyright():
    r"""
    Outputs short license summary for terminal interactions
    """

    return f"""Wulfric Copyright (C) 2023-{datetime.now().year}  Andrey Rybakov
This program comes with ABSOLUTELY NO WARRANTY; for details type 'wulfric warranty'.
This is free software, and you are welcome to redistribute it
under certain conditions; type 'wulfric conditions' for details."""


def _warranty():
    r"""
    Output short warranty summary for terminal interactions
    """

    return f"""THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION."""


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
from calendar import month_name
from datetime import datetime

import numpy as np

from wulfric._decorate_array import print_2d_array
from wulfric.crystal._atoms import get_atom_species
from wulfric.geometry._geometry import absolute_to_relative, get_volume

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def load_poscar(file_object=None, return_comment=False):
    r"""
    Reads crystal structure from the |POSCAR|_ file.

    Parameters
    ----------
    file_object : str of file-like object, optional
        File to be read. If str, then file is opened with the given name.
        Otherwise it has to have ``.readlines()`` method. By default it looks for the
        "POSCAR" file in the current directory. Behaviour for ``str``:

        * Tries to open the file with the name  ``file_object``.
        * Tries to open the file with the name "POSCAR" in the directory ``file_object``.

    return_comment : bool, default False
        Whether to return the comment from the first line of the POSCAR file.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Cell of a crystal structure, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]
    atoms : dict
        Atoms of the crystal structure.
        Positions are always relative to the cell.
    comment : str
        Comment from the first line of the file. If ``return_comment`` is ``True``.

    Examples
    --------

    .. doctest::

        >>> # Load a POSCAR file
        >>> cell, atoms = wulf.io.load_poscar('POSCAR') # doctest: +SKIP
        >>> # It can return the comment from the file as well:
        >>> cell, atoms, comment = wulf.io.load_poscar('POSCAR', return_comment=True) # doctest: +SKIP

    """

    # Open file if needed
    if file_object is None:
        try:
            file_object = open("POSCAR")
        except FileNotFoundError:
            raise FileNotFoundError(
                f'"POSCAR" file not found, looked here: {os.getcwd()}'
            )
    elif isinstance(file_object, str):
        try:
            file_object = open(file_object, "r", encoding="utf-8")
        except FileNotFoundError:
            try:
                file_object = open(
                    os.path.join(file_object, "POSCAR"), "r", encoding="utf-8"
                )
            except FileNotFoundError:
                raise FileNotFoundError(
                    f'"POSCAR" file not found, looked here: {file_object}'
                )

    lines = file_object.readlines()

    comment = lines[0].strip()

    # 1 or 3 numbers
    scale_factor = np.array(list(map(float, lines[1].split())))
    cell = np.array(list(map(lambda x: list(map(float, x.split())), lines[2:5])))
    if len(scale_factor) == 1:
        if scale_factor[0] < 0:
            scale_factor = abs(scale_factor[0] / get_volume(cell))
        cell *= scale_factor
    elif len(scale_factor) == 3 and np.all(scale_factor > 0):
        cell[0] *= scale_factor[0]
        cell[1] *= scale_factor[1]
        cell[2] *= scale_factor[2]
    else:
        raise ValueError(
            "Scale factor has to be a single positive ot negative number or "
            + f"a list of 3 positive numbers, got: {scale_factor}"
        )
    # Read species name and numbers
    species = lines[5].split()
    GOT_SPECIES_NAMES = False
    index = 6
    for i in species:
        try:
            int(i)
        except ValueError:
            GOT_SPECIES_NAMES = True
    if GOT_SPECIES_NAMES:
        species_names = species
        species = lines[6].split()
        index = 7
    else:
        species_names = None
    ions_per_species = list(map(int, species))

    # Skip selective dynamics
    if lines[index][0] in ["S", "s"]:
        index += 1

    # Get mode
    CARTESIAN = False
    if lines[index][0].lower() in ["c", "k"]:
        CARTESIAN = True
    index += 1
    atoms = {"names": [], "positions": []}
    for i in range(len(species)):
        for j in range(ions_per_species[i]):
            coordinates = np.array(list(map(float, lines[index].split()[:3])))
            index += 1
            if CARTESIAN:
                # Both cases (1 or 3 numbers) are covered
                coordinates *= scale_factor
                coordinates = absolute_to_relative(coordinates, cell)
            if species_names is None:
                atoms["names"].append(f"X{i+1}")
                atoms["positions"].append(coordinates)
            else:
                atoms["names"].append(species_names[i])
                atoms["positions"].append(coordinates)

    if return_comment:
        return cell, atoms, comment
    return cell, atoms


def dump_poscar(
    cell,
    atoms,
    file_object="POSCAR",
    comment: str = None,
    decimals=8,
    mode: str = "Direct",
):
    r"""
    Writes crystal structure to the |POSCAR|_ file.

    Parameters
    ----------
    cell : (3, 3) |array-like|_,
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3)
        |array-like|_. Must have either ``"names"`` key with value of ``list`` of ``str``
        of length N or ``"species"`` key with value of ``list`` of ``str`` of length N.
        If ``"species"`` key is not present, try to deduce atom's species from
        ``"names"``, raise error on fail.
    file_object : str of file-like object, optional
        File to be written. If str, then file is opened with the given name.
        Otherwise it has to have ``.write()`` method.
    comment : str, optional
        Comment to be written in the first line of the file. Has to be a single line.
        All new lines symbols are replaced with spaces.
    decimals : int, default 8
        Number of decimals to be written.
    mode : str, default "Direct"
        Mode of the coordinates to be written. Can be "Direct" or "Cartesian".

    Examples
    --------

    .. doctest::

        >>> # Dump a POSCAR file
        >>> wulf.io.dump_poscar(cell, atoms, 'POSCAR') # doctest: +SKIP
        >>> # If you want to write a comment as well:
        >>> wulf.io.dump_poscar(cell, atoms, 'POSCAR', comment='This is a comment') # doctest: +SKIP
        >>> # You can control the amount of decimals in the output:
        >>> wulf.io.dump_poscar(cell, atoms, 'POSCAR', decimals=6) # doctest: +SKIP
        >>> # You can switch the mode of coordinates between 'Cartesian' and 'Direct' (default):
        >>> wulf.io.dump_poscar(cell, atoms, 'POSCAR', mode="Cartesian") # doctest: +SKIP

    """

    cell = np.array(cell, dtype=float)

    # Prepare comment
    if comment is None:
        cd = datetime.now()
        comment = (
            f"Written by wulfric (wulfric.org) "
            f"on {cd.day} {month_name[cd.month]} {cd.year} "
            f"at {cd.hour}:{cd.minute}:{cd.second}"
        )
    else:
        comment = comment.replace("\n", " ")

    # Check mode
    if mode not in ("Direct", "Cartesian"):
        raise ValueError(f'mode has to be "Direct" or "Cartesian", given: {mode}')

    # Prepare atoms
    atoms_list = []
    for i in range(len(atoms["positions"])):
        if "species" in atoms:
            atom_type = atoms["species"][i]
        else:
            atom_type = get_atom_species(atoms["names"][i])
            if atom_type == "X":
                raise ValueError(
                    f"Can not deduce atom's type from the name '{atoms['name'][i]}', while dumping to POSCAR."
                )
        if mode == "Direct":
            atom_position = atoms["positions"][i]
        else:
            atom_position = atoms["positions"][i] @ cell

        atoms_list.append((atom_type, atom_position))

    # Sort atoms by type
    atoms_list = sorted(atoms_list, key=lambda x: x[0])

    # Prepare atom species and coordinates
    atom_species = {}
    atom_coordinates = []
    for atom in atoms_list:
        if atom[0] not in atom_species:
            atom_species[atom[0]] = 1
        else:
            atom_species[atom[0]] += 1
        atom_coordinates.append(atom[1])

    # Open file if needed
    if isinstance(file_object, str):
        file_object = open(file_object, "w", encoding="utf-8")

    # Write
    file_object.write(comment + "\n")
    file_object.write("1.0\n")
    file_object.write(
        print_2d_array(cell, fmt=f".{decimals}f", print_result=False, borders=False)
        + "\n"
    )

    for species in atom_species:
        file_object.write(f"{species} ")
    file_object.write("\n")
    for species in atom_species:
        file_object.write(f"{atom_species[species]} ")
    file_object.write("\n")

    file_object.write(mode + "\n")
    file_object.write(
        print_2d_array(
            atom_coordinates, fmt=f".{decimals}f", print_result=False, borders=False
        )
        + "\n"
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos, sin, tan

import numpy as np

from wulfric.cell._basic_manipulation import get_params, get_reciprocal
from wulfric.cell._lepage import lepage
from wulfric.cell._sc_standardize import get_C_matrix, get_conventional, get_S_matrix
from wulfric.cell._sc_variation import get_variation
from wulfric.constants._numerical import TORADIANS
from wulfric.constants._sc_notation import DEFAULT_K_PATHS, HS_PLOT_NAMES

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _CUB_hs_points():
    r"""
    Get high-symmetry points for the CUB lattice.

    See :ref:`guide_cub` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    return {
        "G": np.array([0, 0, 0]),
        "M": np.array([1 / 2, 1 / 2, 0]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([0, 1 / 2, 0]),
    }


def _FCC_hs_points():
    r"""
    Get high-symmetry points for the FCC lattice.

    See :ref:`guide_fcc` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    return {
        "G": np.array([0, 0, 0]),
        "K": np.array([3 / 8, 3 / 8, 3 / 4]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "U": np.array([5 / 8, 1 / 4, 5 / 8]),
        "W": np.array([1 / 2, 1 / 4, 3 / 4]),
        "X": np.array([1 / 2, 0, 1 / 2]),
    }


def _BCC_hs_points():
    r"""
    Get high-symmetry points for the CUB lattice.

    See :ref:`guide_bcc` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    return {
        "G": np.array([0, 0, 0]),
        "H": np.array([1 / 2, -1 / 2, 1 / 2]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "N": np.array([0, 0, 1 / 2]),
    }


def _TET_hs_points():
    r"""
    Get high-symmetry points for the TET lattice.

    See :ref:`guide_tet` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """
    return {
        "G": np.array([0, 0, 0]),
        "A": np.array([1 / 2, 1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 1 / 2, 0]),
        "R": np.array([0, 1 / 2, 1 / 2]),
        "X": np.array([0, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _BCT_hs_points(variation, conv_a, conv_c):
    r"""
    Get high-symmetry points for the BCT lattice.

    See :ref:`guide_bct` for the details.

    Parameters
    ----------
    variation : str
        BCT variation. Case-insensitive.
    conv_a : float
        Length of the first two lattice vectors of the conventional cell.
    conv_c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    variation = variation.upper()

    if variation == "BCT1":
        eta = (1 + conv_c**2 / conv_a**2) / 4
        kpoints = {
            "G": np.array([0, 0, 0]),
            "M": np.array([-1 / 2, 1 / 2, 1 / 2]),
            "N": np.array([0, 1 / 2, 0]),
            "P": np.array([1 / 4, 1 / 4, 1 / 4]),
            "X": np.array([0, 0, 1 / 2]),
            "Z": np.array([eta, eta, -eta]),
            "Z1": np.array([-eta, 1 - eta, eta]),
        }

    elif variation == "BCT2":
        eta = (1 + conv_a**2 / conv_c**2) / 4
        zeta = conv_a**2 / (2 * conv_c**2)
        kpoints = {
            "G": np.array([0, 0, 0]),
            "N": np.array([0, 1 / 2, 0]),
            "P": np.array([1 / 4, 1 / 4, 1 / 4]),
            "S": np.array([-eta, eta, eta]),
            "S1": np.array([eta, 1 - eta, -eta]),
            "X": np.array([0, 0, 1 / 2]),
            "Y": np.array([-zeta, zeta, 1 / 2]),
            "Y1": np.array([1 / 2, 1 / 2, -zeta]),
            "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
        }
    return kpoints


def _ORC_hs_points():
    r"""
    Get high-symmetry points for the ORC lattice.

    See :ref:`guide_orc` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """
    return {
        "G": np.array([0, 0, 0]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "S": np.array([1 / 2, 1 / 2, 0]),
        "T": np.array([0, 1 / 2, 1 / 2]),
        "U": np.array([1 / 2, 0, 1 / 2]),
        "X": np.array([1 / 2, 0, 0]),
        "Y": np.array([0, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _ORCF_hs_points(variation, conv_a, conv_b, conv_c):
    r"""
    Get high-symmetry points for the ORCF lattice.

    See :ref:`guide_orcf` for the details.

    Parameters
    ----------
    variation : str
        ORCF variation. Case-insensitive.
    conv_a : float
        Length of the first lattice vector of the conventional cell.
    conv_b : float
        Length of the second lattice vector of the conventional cell.
    conv_c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    variation = variation.upper()

    if variation == "ORCF1":
        eta = (1 + conv_a**2 / conv_b**2 + conv_a**2 / conv_c**2) / 4
        zeta = (1 + conv_a**2 / conv_b**2 - conv_a**2 / conv_c**2) / 4

        kpoints = {
            "G": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2 + zeta, zeta]),
            "A1": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "T": np.array([1, 1 / 2, 1 / 2]),
            "X": np.array([0, eta, eta]),
            "X1": np.array([1, 1 - eta, 1 - eta]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }
    elif variation == "ORCF2":
        eta = (1 + conv_a**2 / conv_b**2 - conv_a**2 / conv_c**2) / 4
        delta = (1 + conv_b**2 / conv_a**2 - conv_b**2 / conv_c**2) / 4
        phi = (1 + conv_c**2 / conv_b**2 - conv_c**2 / conv_a**2) / 4

        kpoints = {
            "G": np.array([0, 0, 0]),
            "C": np.array([1 / 2, 1 / 2 - eta, 1 - eta]),
            "C1": np.array([1 / 2, 1 / 2 + eta, eta]),
            "D": np.array([1 / 2 - delta, 1 / 2, 1 - delta]),
            "D1": np.array([1 / 2 + delta, 1 / 2, delta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "H": np.array([1 - phi, 1 / 2 - phi, 1 / 2]),
            "H1": np.array([phi, 1 / 2 + phi, 1 / 2]),
            "X": np.array([0, 1 / 2, 1 / 2]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }

    elif variation == "ORCF3":
        eta = (1 + conv_a**2 / conv_b**2 + conv_a**2 / conv_c**2) / 4
        zeta = (1 + conv_a**2 / conv_b**2 - conv_a**2 / conv_c**2) / 4

        kpoints = {
            "G": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2 + zeta, zeta]),
            "A1": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "T": np.array([1, 1 / 2, 1 / 2]),
            "X": np.array([0, eta, eta]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }
    return kpoints


def _ORCI_hs_points(conv_a, conv_b, conv_c):
    r"""
    Get high-symmetry points for the ORCI lattice.

    See :ref:`guide_orci` for the details.

    Parameters
    ----------
    conv_a : float
        Length of the first lattice vector of the conventional cell.
    conv_b : float
        Length of the second lattice vector of the conventional cell.
    conv_c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    zeta = (1 + conv_a**2 / conv_c**2) / 4
    eta = (1 + conv_b**2 / conv_c**2) / 4
    delta = (conv_b**2 - conv_a**2) / (4 * conv_c**2)
    mu = (conv_a**2 + conv_b**2) / (4 * conv_c**2)

    return {
        "G": np.array([0, 0, 0]),
        "L": np.array([-mu, mu, 1 / 2 - delta]),
        "L1": np.array([mu, -mu, 1 / 2 + delta]),
        "L2": np.array([1 / 2 - delta, 1 / 2 + delta, -mu]),
        "R": np.array([0, 1 / 2, 0]),
        "S": np.array([1 / 2, 0, 0]),
        "T": np.array([0, 0, 1 / 2]),
        "W": np.array([1 / 4, 1 / 4, 1 / 4]),
        "X": np.array([-zeta, zeta, zeta]),
        "X1": np.array([zeta, 1 - zeta, -zeta]),
        "Y": np.array([eta, -eta, eta]),
        "Y1": np.array([1 - eta, eta, -eta]),
        "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
    }


def _ORCC_hs_points(conv_a, conv_b):
    r"""
    Get high-symmetry points for the ORCC lattice.

    See :ref:`guide_orcc` for the details.

    Parameters
    ----------
    conv_a : float
        Length of the first lattice vector of the conventional cell.
    conv_b : float
        Length of the second lattice vector of the conventional cell.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    zeta = (1 + conv_a**2 / conv_b**2) / 4

    return {
        "G": np.array([0, 0, 0]),
        "A": np.array([zeta, zeta, 1 / 2]),
        "A1": np.array([-zeta, 1 - zeta, 1 / 2]),
        "R": np.array([0, 1 / 2, 1 / 2]),
        "S": np.array([0, 1 / 2, 0]),
        "T": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([zeta, zeta, 0]),
        "X1": np.array([-zeta, 1 - zeta, 0]),
        "Y": np.array([-1 / 2, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _HEX_hs_points():
    r"""
    Get high-symmetry points for the HEX lattice.

    See :ref:`guide_hex` for the details.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    return {
        "G": np.array([0, 0, 0]),
        "A": np.array([0, 0, 1 / 2]),
        "H": np.array([1 / 3, 1 / 3, 1 / 2]),
        "K": np.array([1 / 3, 1 / 3, 0]),
        "L": np.array([1 / 2, 0, 1 / 2]),
        "M": np.array([1 / 2, 0, 0]),
    }


def _RHL_hs_points(variation, conv_alpha):
    r"""
    Get high-symmetry points for the RHL lattice.

    See :ref:`guide_rhl` for the details.

    Parameters
    ----------
    variation : str
        RHL variation. Case-insensitive.
    alpha : float
        Angle between the lattice vectors.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    variation = variation.upper()

    conv_alpha *= TORADIANS

    if variation == "RHL1":
        eta = (1 + 4 * cos(conv_alpha)) / (2 + 4 * cos(conv_alpha))
        nu = 3 / 4 - eta / 2

        return {
            "G": np.array([0, 0, 0]),
            "B": np.array([eta, 1 / 2, 1 - eta]),
            "B1": np.array([1 / 2, 1 - eta, eta - 1]),
            "F": np.array([1 / 2, 1 / 2, 0]),
            "L": np.array([1 / 2, 0, 0]),
            "L1": np.array([0, 0, -1 / 2]),
            "P": np.array([eta, nu, nu]),
            "P1": np.array([1 - nu, 1 - nu, 1 - eta]),
            "P2": np.array([nu, nu, eta - 1]),
            "Q": np.array([1 - nu, nu, 0]),
            "X": np.array([nu, 0, -nu]),
            "Z": np.array([1 / 2, 1 / 2, 1 / 2]),
        }

    elif variation == "RHL2":
        eta = 1 / (2 * tan(conv_alpha / 2) ** 2)
        nu = 3 / 4 - eta / 2

        return {
            "G": np.array([0, 0, 0]),
            "F": np.array([1 / 2, -1 / 2, 0]),
            "L": np.array([1 / 2, 0, 0]),
            "P": np.array([1 - nu, -nu, 1 - nu]),
            "P1": np.array([nu, nu - 1, nu - 1]),
            "Q": np.array([eta, eta, eta]),
            "Q1": np.array([1 - eta, -eta, -eta]),
            "Z": np.array([1 / 2, -1 / 2, 1 / 2]),
        }


def _MCL_hs_points(conv_b, conv_c, conv_alpha):
    r"""
    Get high-symmetry points for the MCL lattice.

    See :ref:`guide_mcl` for the details.

    Parameters
    ----------
    conv_b : float
        Length of the second lattice vector of the conventional cell.
    conv_c : float
        Length of the third lattice vector of the conventional cell.
    conv_alpha : float
        Angle between the lattice vectors.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """
    conv_alpha *= TORADIANS

    eta = (1 - conv_b * cos(conv_alpha) / conv_c) / (2 * sin(conv_alpha) ** 2)
    nu = 1 / 2 - eta * conv_c * cos(conv_alpha) / conv_b

    return {
        "G": np.array([0, 0, 0]),
        "A": np.array([1 / 2, 1 / 2, 0]),
        "C": np.array([0, 1 / 2, 1 / 2]),
        "D": np.array([1 / 2, 0, 1 / 2]),
        "D1": np.array([1 / 2, 0, -1 / 2]),
        "E": np.array([1 / 2, 1 / 2, 1 / 2]),
        "H": np.array([0, eta, 1 - nu]),
        "H1": np.array([0, 1 - eta, nu]),
        "H2": np.array([0, eta, -nu]),
        "M": np.array([1 / 2, eta, 1 - nu]),
        "M1": np.array([1 / 2, 1 - eta, nu]),
        "M2": np.array([1 / 2, eta, -nu]),
        "X": np.array([0, 1 / 2, 0]),
        "Y": np.array([0, 0, 1 / 2]),
        "Y1": np.array([0, 0, -1 / 2]),
        "Z": np.array([1 / 2, 0, 0]),
    }


def _MCLC_hs_points(variation, conv_a, conv_b, conv_c, conv_alpha):
    r"""
    Get high-symmetry points for the MCLC lattice.

    See :ref:`guide_mclc` for the details.

    Parameters
    ----------
    variation : str
        MCLC variation.  Case-insensitive.
    conv_a : float
        Length of the first lattice vector of the conventional cell.
    conv_b : float
        Length of the second lattice vector of the conventional cell.
    conv_c : float
        Length of the third lattice vector of the conventional cell.
    conv_alpha : float
        Angle between the lattice vectors.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """
    variation = variation.upper()

    conv_alpha *= TORADIANS
    # Parameters
    if variation in ["MCLC1", "MCLC2"]:
        zeta = (2 - conv_b * cos(conv_alpha) / conv_c) / (4 * sin(conv_alpha) ** 2)
        eta = 1 / 2 + 2 * zeta * conv_c * cos(conv_alpha) / conv_b
        psi = 3 / 4 - conv_a**2 / (4 * conv_b**2 * sin(conv_alpha) ** 2)
        phi = psi + (3 / 4 - psi) * conv_b * cos(conv_alpha) / conv_c
    elif variation in ["MCLC3", "MCLC4"]:
        mu = (1 + conv_b**2 / conv_a**2) / 4
        delta = conv_b * conv_c * cos(conv_alpha) / (2 * conv_a**2)
        zeta = (
            mu
            - 1 / 4
            + (1 - conv_b * cos(conv_alpha) / conv_c) / (4 * sin(conv_alpha) ** 2)
        )
        eta = 1 / 2 + 2 * zeta * conv_c * cos(conv_alpha) / conv_b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta
    elif variation == "MCLC5":
        zeta = (
            conv_b**2 / conv_a**2
            + (1 - conv_b * cos(conv_alpha) / conv_c) / sin(conv_alpha) ** 2
        ) / 4
        eta = 1 / 2 + 2 * zeta * conv_c * cos(conv_alpha) / conv_b
        mu = (
            eta / 2
            + conv_b**2 / (4 * conv_a**2)
            - conv_b * conv_c * cos(conv_alpha) / (2 * conv_a**2)
        )
        nu = 2 * mu - zeta
        rho = 1 - zeta * conv_a**2 / conv_b**2
        omega = (
            (4 * nu - 1 - conv_b**2 * sin(conv_alpha) ** 2 / conv_a**2)
            * conv_c
            / (2 * conv_b * cos(conv_alpha))
        )
        delta = zeta * conv_c * cos(conv_alpha) / conv_b + omega / 2 - 1 / 4

    # Path
    if variation == "MCLC1":
        return {
            "G": np.array([0, 0, 0]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F1": np.array([zeta, zeta, eta]),
            "F2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 1 / 2]),
            "I1": np.array([1 - phi, phi - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "X": np.array([1 - psi, psi - 1, 0]),
            "X1": np.array([psi, 1 - psi, 0]),
            "X2": np.array([psi - 1, -psi, 0]),
            "Y": np.array([1 / 2, 1 / 2, 0]),
            "Y1": np.array([-1 / 2, -1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }
    elif variation == "MCLC2":
        return {
            "G": np.array([0, 0, 0]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F1": np.array([zeta, zeta, eta]),
            "F2": np.array([-zeta, -zeta, 1 - eta]),
            "F3": np.array([1 - zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 1 / 2]),
            "I1": np.array([1 - phi, phi - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "X": np.array([1 - psi, psi - 1, 0]),
            "Y": np.array([1 / 2, 1 / 2, 0]),
            "Y1": np.array([-1 / 2, -1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }
    elif variation == "MCLC3":
        return {
            "G": np.array([0, 0, 0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "F1": np.array([phi, phi - 1, psi]),
            "F2": np.array([1 - phi, -phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([1 / 2, -1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }
    elif variation == "MCLC4":
        return {
            "G": np.array([0, 0, 0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([1 / 2, -1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }
    elif variation == "MCLC5":
        return {
            "G": np.array([0, 0, 0]),
            "F": np.array([nu, nu, omega]),
            "F1": np.array([1 - nu, 1 - nu, 1 - omega]),
            "F2": np.array([nu, nu - 1, omega]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([rho, 1 - rho, 1 / 2]),
            "I1": np.array([1 - rho, rho - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }


def _TRI_hs_points(variation):
    r"""
    Get high-symmetry points for the TRI lattice.

    See :ref:`guide_tri` for the details.

    Parameters
    ----------
    variation : str
        TRI variation. Case-insensitive.

    Returns
    -------
    kpoints : dict
        High-symmetry points.
    """

    variation = variation.upper()

    if variation in ["TRI1A", "TRI2A"]:
        return {
            "G": np.array([0, 0, 0]),
            "L": np.array([1 / 2, 1 / 2, 0]),
            "M": np.array([0, 1 / 2, 1 / 2]),
            "N": np.array([1 / 2, 0, 1 / 2]),
            "R": np.array([1 / 2, 1 / 2, 1 / 2]),
            "X": np.array([1 / 2, 0, 0]),
            "Y": np.array([0, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

    elif variation in ["TRI1B", "TRI2B"]:
        return {
            "G": np.array([0, 0, 0]),
            "L": np.array([1 / 2, -1 / 2, 0]),
            "M": np.array([0, 0, 1 / 2]),
            "N": np.array([-1 / 2, -1 / 2, 1 / 2]),
            "R": np.array([0, -1 / 2, 1 / 2]),
            "X": np.array([0, -1 / 2, 0]),
            "Y": np.array([1 / 2, 0, 0]),
            "Z": np.array([-1 / 2, 0, 1 / 2]),
        }


def get_hs_data(
    cell,
    return_relative=True,
    lattice_type=None,
    lattice_variation=None,
    S_matrix=None,
    C_matrix=None,
    length_tolerance=1e-8,
    angle_tolerance=1e-4,
):
    r"""
    Return information about high symmetry points and path as defined in the paper by
    Setyawan and Curtarolo [1]_.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    return_relative : bool, default True
        Whether to return coordinates as relative to the reciprocal cell or in absolute
        coordinates in the reciprocal Cartesian space.
    lattice_type : str, optional
        One of the 14 lattice types that correspond to the provided ``cell``,
        case-insensitive. If not provided, then computed automatically from ``cell``. If
        provided, then it user's responsibility to ensure that ``lattice_type`` is
        correct.
    lattice_variation : str, optional
        One of the lattice variations that correspond to the provided ``cell`` and
        ``lattice_type``. If not provided, then computed automatically. Case-insensitive.
    S_matrix : (3, 3) |array-like|_, optional
        Transformation matrix S. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    C_matrix : (3, 3) |array-like|_, optional
        Transformation matrix C. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors).  Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    coordinates : list of (3, 3) :numpy:`ndarray`
        Coordinates of the high symmetry points in reciprocal space. Relative to the
        reciprocal cell.
    names: list of str
        Names of the high symmetry points. Used for programming, not for plotting. Have
        the same length as ``coordinates``.
    labels : list of str
        List of the high symmetry points labels for plotting. Have the same length as
        ``coordinates``. Labels are not necessary equal to the names.
    path : str
        K path. High symmetry points are referenced by elements of ``names``.

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    See Also
    --------
    wulfric.Kpoints : Class with a convenient interface for the same information.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = wulf.cell.get_cell_example("hex")
        >>> coordinates, names, labels, kpath = wulf.cell.get_hs_data(cell)
        >>> kpath
        'G-M-K-G-A-L-H-A|L-M|K-H'
        >>> labels
        ['$\\Gamma$', 'A', 'H', 'K', 'L', 'M']
        >>> names
        ['G', 'A', 'H', 'K', 'L', 'M']
        >>> coordinates
        [array([0., 0., 0.]), array([0. , 0. , 0.5]), array([0.33333333, 0.33333333, 0.5       ]), array([0.33333333, 0.33333333, 0.        ]), array([0.5, 0. , 0.5]), array([0.5, 0. , 0. ])]

    """

    cell = np.array(cell, dtype=float)

    if lattice_type is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

    lattice_type = lattice_type.upper()

    if lattice_variation is None:
        lattice_variation = get_variation(
            cell=cell, lattice_type=lattice_type, angle_tolerance=angle_tolerance
        )

    lattice_variation = lattice_variation.upper()

    if C_matrix is None:
        C_matrix = get_C_matrix(lattice_type)
    else:
        C_matrix = np.array(C_matrix, dtype=float)

    if S_matrix is None:
        S_matrix = get_S_matrix(
            cell,
            lattice_type,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )
    else:
        S_matrix = np.array(S_matrix, dtype=float)

    if lattice_type in ["BCT", "ORCF", "ORCI", "ORCC", "RHL", "MCL", "MCLC"]:
        conv_a, conv_b, conv_c, conv_alpha, conv_beta, conv_gamma = get_params(
            get_conventional(cell, S_matrix=S_matrix, C_matrix=C_matrix)
        )

    if lattice_type == "CUB":
        hs_points = _CUB_hs_points()
    elif lattice_type == "FCC":
        hs_points = _FCC_hs_points()
    elif lattice_type == "BCC":
        hs_points = _BCC_hs_points()
    elif lattice_type == "TET":
        hs_points = _TET_hs_points()
    elif lattice_type == "BCT":
        hs_points = _BCT_hs_points(lattice_variation, conv_a, conv_c)
    elif lattice_type == "ORC":
        hs_points = _ORC_hs_points()
    elif lattice_type == "ORCF":
        hs_points = _ORCF_hs_points(lattice_variation, conv_a, conv_b, conv_c)
    elif lattice_type == "ORCI":
        hs_points = _ORCI_hs_points(conv_a, conv_b, conv_c)
    elif lattice_type == "ORCC":
        hs_points = _ORCC_hs_points(conv_a, conv_b)
    elif lattice_type == "HEX":
        hs_points = _HEX_hs_points()
    elif lattice_type == "RHL":
        hs_points = _RHL_hs_points(lattice_variation, conv_alpha)
    elif lattice_type == "MCL":
        hs_points = _MCL_hs_points(conv_b, conv_c, conv_alpha)
    elif lattice_type == "MCLC":
        hs_points = _MCLC_hs_points(
            lattice_variation, conv_a, conv_b, conv_c, conv_alpha
        )
    elif lattice_type == "TRI":
        hs_points = _TRI_hs_points(lattice_variation)

    names = []
    labels = []
    coordinates = []

    for point in hs_points:
        names.append(point)
        # Compute relative coordinates with respect to the
        # non-standardized primitive cell
        # here hs_points[point] <- \tilde{g} and coordinates <- g
        coordinates.append(np.linalg.inv(S_matrix).T @ hs_points[point])

        # Post-process two edge cases
        if point == "S" and lattice_type == "BCT":
            labels.append("$\\Sigma$")
        elif point == "S1" and lattice_type == "BCT":
            labels.append("$\\Sigma_1$")
        # General assignment
        else:
            labels.append(HS_PLOT_NAMES[point])

    if not return_relative:

        rcell = get_reciprocal(cell)

        for i in range(len(coordinates)):
            coordinates[i] = coordinates[i] @ rcell

    return coordinates, names, labels, DEFAULT_K_PATHS[lattice_variation]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from wulfric._exceptions import FailedToDeduceAtomSpecies
from wulfric.constants._atoms import ATOM_SPECIES

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_atom_species(name: str, raise_on_fail=False) -> str:
    r"""
    Attempts to identify atom's type based on its name (i.e. Cr1 -> Cr, ...).

    If no type is identified, then return "X".

    Parameters
    ----------
    name : str
        Name of the atom.
    raise_on_fail : bool, default False
        Whether to raise an exception if automatic species deduction fails.

    Returns
    -------
    species : str
        Species of the atom.

    Raises
    ------
    FailedToDeduceAtomSpecies
        If ``raise_on_fail = True`` and automatic species deduction fails.

    Warnings
    --------
    If ``raise_on_fail = True`` and automatic species deduction fails, then
    ``RuntimeWarning`` is issued, and atom species is set to "X".

    Examples
    --------

    .. doctest::

        >>> from wulfric.crystal import get_atom_species
        >>> get_atom_species("@%^#$")
        'X'
        >>> get_atom_species("Cr")
        'Cr'
        >>> get_atom_species("Cr1")
        'Cr'
        >>> get_atom_species("_3341Cr")
        'Cr'
        >>> get_atom_species("cr")
        'Cr'
        >>> get_atom_species("S")
        'S'
        >>> get_atom_species("Se")
        'Se'
        >>> get_atom_species("Sp")
        'S'
        >>> get_atom_species("123a")
        'X'
        >>> get_atom_species("CrSBr")
        'Cr'

    Notes
    -----
    If ``name`` contains several possible atom types of length 2
    as substrings, then the type is equal to the first one found.
    """

    atom_type = "X"
    for trial_type in ATOM_SPECIES:
        if trial_type.lower() in name.lower():
            atom_type = trial_type
            # Maximum amount of characters in the atom type
            # Some 1-character types are parts of some 2-character types (i.e. "Se" and "S")
            # If type of two characters is found then it is unique,
            # If type of one character is found, then the search must continue
            if len(atom_type) == 2:
                break

    if atom_type == "X":
        if raise_on_fail:
            raise FailedToDeduceAtomSpecies(name=name)
        else:
            import warnings

            warnings.warn(
                f"Atom species deduction failed for '{name}'. Set species to 'X'",
                RuntimeWarning,
            )

    return atom_type


def populate_atom_species(atoms, raise_on_fail=False) -> None:
    r"""
    Populate atom species, based on their names.
    If atom species are already present in the ``atoms``, then they will be overwritten.

    Parameters
    ----------
    atoms : dict
        Dictionary with atoms. Must have a ``names`` with the value of ``list`` of N
        ``str``.
    raise_on_fail : bool, default False
        Whether to raise an error if the atom type can not be deduced based on its name.

    Raises
    ------
    FailedToDeduceAtomSpecies
        If ``raise_on_fail = True`` and automatic species deduction fails.

    Warnings
    --------
    If ``raise_on_fail = True`` and automatic species deduction fails, then
    ``RuntimeWarning`` is issued, and atom species is set to "X".

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {"names" : ["Cr1", "cr2", "Br3", "S4", "fe5", "Fe6"]}
        >>> atoms
        {'names': ['Cr1', 'cr2', 'Br3', 'S4', 'fe5', 'Fe6']}
        >>> wulf.crystal.populate_atom_species(atoms)
        >>> atoms
        {'names': ['Cr1', 'cr2', 'Br3', 'S4', 'fe5', 'Fe6'], 'species': ['Cr', 'Cr', 'Br', 'S', 'Fe', 'Fe']}

    """

    atoms["species"] = []

    for i in range(len(atoms["names"])):
        atoms["species"].append(
            get_atom_species(atoms["names"][i], raise_on_fail=raise_on_fail)
        )


def ensure_unique_names(atoms, strategy: str = "all") -> None:
    r"""
    Ensures that atoms have unique ``"names"``.

    If atom names are already unique, then this function does nothing.

    .. versionadded:: 0.5.1

    Parameters
    ----------
    atoms : dict
        Dictionary with atoms. Must have a ``"names"`` keyword with the value of
        ``list`` of N ``str``.
    strategy : str, default "all"
        Strategy for the modification of atom names. Supported strategies are

        * "all"

          Add an index to the end of every atom, starting from 1.
        * "repeated-only"

          Add an index only to the repeated names, index starts with 1, independently for
          each repeated grooup. (See examples)

        Case-insensitive.

    Raises
    ------
    ValueError
        If ``strategy`` is not supported.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {"names" : ["Cr1", "Cr2", "Br", "Br", "S", "S"]}
        >>> # Default strategy is "all"
        >>> wulf.crystal.ensure_unique_names(atoms)
        >>> atoms
        {'names': ['Cr11', 'Cr22', 'Br3', 'Br4', 'S5', 'S6']}
        >>> atoms = {"names" : ["Cr1", "Cr2", "Br", "Br", "S", "S"]}
        >>> wulf.crystal.ensure_unique_names(atoms, strategy="repeated-only")
        >>> atoms
        {'names': ['Cr1', 'Cr2', 'Br1', 'Br2', 'S1', 'S2']}
        >>> # Nothing happens if atom names are already unique
        >>> wulf.crystal.ensure_unique_names(atoms)
        >>> atoms
        {'names': ['Cr1', 'Cr2', 'Br1', 'Br2', 'S1', 'S2']}
        >>> wulf.crystal.ensure_unique_names(atoms, strategy="repeated-only")
        >>> atoms
        {'names': ['Cr1', 'Cr2', 'Br1', 'Br2', 'S1', 'S2']}

    """

    SUPPORTED_STRATEGIES = ["all", "repeated-only"]
    strategy = strategy.lower()

    if strategy not in SUPPORTED_STRATEGIES:
        raise ValueError(
            f"{strategy} strategy is not supported. Supported are:\n"
            + ("\n").join([f"  * {i}" for i in SUPPORTED_STRATEGIES])
        )

    names_unique = len(atoms["names"]) == len(set(atoms["names"]))

    if not names_unique and strategy == "all":

        for i in range(len(atoms["names"])):
            atoms["names"][i] += f"{i + 1}"

    if not names_unique and strategy == "repeated-only":
        counter = {}
        for name in atoms["names"]:
            if name not in counter:
                counter[name] = [1, 1]
            else:
                counter[name][1] += 1

        for i in range(len(atoms["names"])):
            name = atoms["names"][i]
            total = counter[name][1]
            met = counter[name][0]
            if total > 1:
                atoms["names"][i] += str(counter[name][0])
                counter[name][0] += 1


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


old_dir = set(dir())
old_dir.add("old_dir")


class StandardizationTypeMismatch(Exception):
    r"""
    Raised if standardization functions is called on the cell that does not match the
    expected lattice type. It is not guaranteed to be raised if the lattice type is
    incorrect.
    """

    def __init__(self, expected_lattice_type, step=None):
        if step is not None:
            self.message = f"{step} step of the standardization process fails. "
        else:
            self.message = ""
        self.message += f"Are you sure that the cell is {expected_lattice_type}?"

    def __str__(self):
        return self.message


class FailedToDeduceAtomSpecies(Exception):
    r"""
    Raised when the automatic deduction of the atom species from its name fails.
    """

    def __init__(self, name: str):
        self.message = f"Tried to deduce name from '{name}'. Failed."

    def __str__(self):
        return self.message


class NiggliReductionFailed(Exception):
    r"""
    Raised when niggli reduction reaches ``max_iterations``.
    """

    def __init__(self, max_iterations: int):
        self.message = f"Niggli reduction algorithm reached maximum amount of iterations: {max_iterations}"

    def __str__(self):
        return self.message


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
